import os
import random
import gc
import copy
import traceback
import numpy as np
import pandas as pd
import torch
import time
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import DataLoader, TensorDataset
from torch.optim.lr_scheduler import OneCycleLR

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score

import cv2
import matplotlib.pyplot as plt

# TINTO library imports
from TINTOlib.tinto import TINTO
from TINTOlib.supertml import SuperTML
from TINTOlib.igtd import IGTD
from TINTOlib.refined import REFINED
from TINTOlib.barGraph import BarGraph
from TINTOlib.distanceMatrix import DistanceMatrix
from TINTOlib.combination import Combination
from TINTOlib.featureWrap import FeatureWrap
from TINTOlib.bie import BIE

import warnings
warnings.filterwarnings("ignore", category=UserWarning)


# ------------------------- MODEL DEFINITIONS ------------------------- #
from vit_pytorch.vit import ViT

def find_divisors(n):
    divisors = []
    for i in range(1, int(n**0.5) + 1):
        if n % i == 0:
            divisors.append(i)
            if i != n // i:  # Check to include both divisors if they are not the same
                divisors.append(n // i)
    divisors.sort()
    return divisors

class Model1(nn.Module):
    def __init__(self, imgs_shape, patch_size):
        super(Model1, self).__init__()
        
        # ViT branch
        self.vit = ViT(
            image_size = imgs_shape,
            patch_size = patch_size,
            dim = 32,
            depth = 2,
            heads = 4,
            mlp_dim = 64,
            dropout = 0.0,
            emb_dropout = 0.0
        )
        
        self.mlp = nn.Sequential(
            nn.Linear(32, 16),
            nn.ReLU(),
            nn.Linear(16, 8),
            nn.ReLU(),
            nn.Linear(8, 1)
        )

    def forward(self, vit_input):
        vit_output = self.vit(vit_input)
        mlp_output = self.mlp(vit_output)
        return mlp_output


class Model2(nn.Module):
    def __init__(self, imgs_shape, patch_size):
        super(Model2, self).__init__()
        
        # Enhanced ViT branch with increased depth and heads
        self.vit = ViT(
            image_size = imgs_shape,
            patch_size = patch_size,
            dim = 128,  # Increased dimensionality
            depth = 4,  # More layers
            heads = 8,  # More attention heads
            mlp_dim = 256,
            dropout = 0.0,
            emb_dropout = 0.0
        )
        
        # More complex MLP
        self.mlp = nn.Sequential(
            nn.Linear(128, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            nn.Linear(32, 16),
            nn.ReLU(),
            nn.Linear(16, 1)
        )

    def forward(self, vit_input):
        vit_output = self.vit(vit_input)
        mlp_output = self.mlp(vit_output)
        return mlp_output


# ------------------------- HELPER FUNCTIONS ------------------------- #

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from torch.utils.data import DataLoader, TensorDataset
import os
import cv2

# Function to load and preprocess data
def load_and_preprocess_data(df, images_folder, image_model, problem_type, batch_size=32, seed=64):
    
    X_train, X_val = train_test_split(df, test_size=0.20, random_state=seed)
    X_val, X_test = train_test_split(X_val, test_size=0.50, random_state=seed)
    X_train = X_train.reset_index(drop=True)
    X_val = X_val.reset_index(drop=True)
    X_test = X_test.reset_index(drop=True)
    
    ### X_train
    # Generate the images if the folder does not exist
    if not os.path.exists(f'{images_folder}/train'):
        #Generate thet images
        image_model.fit_transform(X_train, f'{images_folder}/train')
        image_model.saveHyperparameters(f'{images_folder}'+"/model.pkl")
    else:
        print("The images are already generated")

    img_paths = os.path.join(f'{images_folder}/train',problem_type+".csv")

    print(img_paths)
    
    imgs = pd.read_csv(img_paths)
    
    # Update image paths
    imgs["images"] = images_folder + "/train/" + imgs["images"]

    # Combine datasets
    combined_dataset = pd.concat([imgs, X_train], axis=1)

    # Split data
    X_train = combined_dataset.drop(df.columns[-1], axis=1).drop("values", axis=1)
    y_train = combined_dataset["values"]
        
    ### X_val
    # Generate the images if the folder does not exist
    if not os.path.exists(f'{images_folder}/val'):
        #Generate thet images
        image_model.transform(X_val, f'{images_folder}/val')
    else:
        print("The images are already generated")

    img_paths = os.path.join(f'{images_folder}/val',problem_type+".csv")

    print(img_paths)
    
    imgs = pd.read_csv(img_paths)

    # Update image paths
    imgs["images"] = images_folder + "/val/" + imgs["images"]

    # Combine datasets
    combined_dataset = pd.concat([imgs, X_val], axis=1)

    # Split data
    X_val = combined_dataset.drop(df.columns[-1], axis=1).drop("values", axis=1)
    y_val = combined_dataset["values"]
    
    ### X_test
    # Generate the images if the folder does not exist
    if not os.path.exists(f'{images_folder}/test'):
        #Generate thet images
        image_model.transform(X_test, f'{images_folder}/test')
    else:
        print("The images are already generated")

    img_paths = os.path.join(f'{images_folder}/test',problem_type+".csv")

    print(img_paths)
    
    imgs = pd.read_csv(img_paths)

    # Update image paths
    imgs["images"] = images_folder + "/test/" + imgs["images"]

    # Combine datasets
    combined_dataset = pd.concat([imgs, X_test], axis=1)

    # Split data
    X_test = combined_dataset.drop(df.columns[-1], axis=1).drop("values", axis=1)
    y_test = combined_dataset["values"]
    
    # Numerical data
    X_train_num = X_train.drop("images", axis=1)
    X_val_num = X_val.drop("images", axis=1)
    X_test_num = X_test.drop("images", axis=1)

    # Image data
    X_train_img = np.array([cv2.imread(img) for img in X_train["images"]])
    X_val_img = np.array([cv2.imread(img) for img in X_val["images"]])
    X_test_img = np.array([cv2.imread(img) for img in X_test["images"]])

    ## Create a MinMaxScaler object
    scaler = MinMaxScaler()
#
    ## Scale numerical data
    X_train_num = pd.DataFrame(scaler.fit_transform(X_train_num), columns=X_train_num.columns)
    X_val_num = pd.DataFrame(scaler.transform(X_val_num), columns=X_val_num.columns)
    X_test_num = pd.DataFrame(scaler.transform(X_test_num), columns=X_test_num.columns)

    attributes = len(X_train_num.columns)
    height, width, channels = X_train_img[0].shape
    imgs_shape = (channels, height, width)

    print("Images shape: ", imgs_shape)
    print("Attributes: ", attributes)
    # Convert data to PyTorch tensors
    X_train_num_tensor = torch.as_tensor(X_train_num.values, dtype=torch.float32)
    X_val_num_tensor = torch.as_tensor(X_val_num.values, dtype=torch.float32)
    X_test_num_tensor = torch.as_tensor(X_test_num.values, dtype=torch.float32)
    X_train_img_tensor = torch.as_tensor(X_train_img, dtype=torch.float32).permute(0, 3, 1, 2)
    X_val_img_tensor = torch.as_tensor(X_val_img, dtype=torch.float32).permute(0, 3, 1, 2)
    X_test_img_tensor = torch.as_tensor(X_test_img, dtype=torch.float32).permute(0, 3, 1, 2)
    y_train_tensor = torch.as_tensor(y_train.values, dtype=torch.float32).reshape(-1, 1)
    y_val_tensor = torch.as_tensor(y_val.values, dtype=torch.float32).reshape(-1, 1)
    y_test_tensor = torch.as_tensor(y_test.values, dtype=torch.float32).reshape(-1, 1)

    # Normalize images to [0, 1]
    #X_train_img_tensor = X_train_img_tensor / 255.0
    #X_val_img_tensor = X_val_img_tensor / 255.0
    #X_test_img_tensor = X_test_img_tensor / 255.0

    # Create DataLoaders
    train_dataset = TensorDataset(X_train_img_tensor, y_train_tensor)
    val_dataset = TensorDataset(X_val_img_tensor, y_val_tensor)
    test_dataset = TensorDataset(X_test_img_tensor, y_test_tensor)

    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, pin_memory=True)
    val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False, pin_memory=True)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False, pin_memory=True)

    return train_loader, val_loader, test_loader, attributes, imgs_shape 


def compile_and_fit(model, train_loader, val_loader, test_loader, dataset_name, model_name, batch_size=32, epochs=200, min_lr=1e-3, max_lr=1, device='cuda', weight_decay=1e-2):
    model = model.to(device)
    loss_fn = nn.MSELoss()
    optimizer = optim.AdamW(model.parameters(), lr=min_lr, weight_decay=weight_decay)
    
    total_steps = epochs * len(train_loader)
    scheduler = OneCycleLR(optimizer, max_lr=max_lr, div_factor=max_lr/min_lr, total_steps=total_steps, pct_start=0.3, final_div_factor=1)
    
    best_val_loss = float('inf')
    early_stopping_counter = 0
    early_stopping_patience = 20
    best_model = None
    best_epoch = 0
    warm_up_epochs = epochs*0.3

    history = {'train_loss': [], 'val_loss': [], 'train_mse': [], 'val_mse': [], 'train_rmse': [], 'val_rmse': [], 'learning_rate': [], 'epoch_time': []}

    start_time = time.time()
    
    for epoch in range(epochs):
        epoch_start_time = time.time()

        model.train()
        train_loss = 0.0
        train_predictions = []
        train_targets = []
        for img_data, targets in train_loader:
            img_data, targets = img_data.to(device, non_blocking=True), targets.to(device, non_blocking=True)
            
            optimizer.zero_grad()
            outputs = model(img_data)
            loss = loss_fn(outputs, targets)
            loss.backward()
            optimizer.step()
            scheduler.step()
            
            train_loss += loss.item()
            train_predictions.extend(outputs.cpu().detach().numpy())
            train_targets.extend(targets.cpu().numpy())

        model.eval()
        val_loss = 0.0
        val_predictions = []
        val_targets = []
        with torch.no_grad():
            for img_data, targets in val_loader:
                img_data, targets = img_data.to(device, non_blocking=True), targets.to(device, non_blocking=True)
                outputs = model(img_data)
                loss = loss_fn(outputs, targets)
                
                val_loss += loss.item()
                val_predictions.extend(outputs.cpu().numpy())
                val_targets.extend(targets.cpu().numpy())

        train_loss /= len(train_loader)
        val_loss /= len(val_loader)
        
        # Get the current learning rate
        current_lr = scheduler.get_last_lr()
        
        
        if val_loss < best_val_loss:
            best_val_loss = val_loss
            best_model = copy.deepcopy(model.state_dict())
            best_epoch = epoch + 1
            #early_stopping_counter = 0
        #else:
            #if epoch > warm_up_epochs:
                #early_stopping_counter += 1
                #if early_stopping_counter >= early_stopping_patience:
                    #print(f"Early stopping triggered at epoch {epoch+1}")
                    #break

        train_mse = mean_squared_error(train_targets, train_predictions)
        train_rmse = np.sqrt(train_mse)
        val_mse = mean_squared_error(val_targets, val_predictions)
        val_rmse = np.sqrt(val_mse)
        train_r2 = r2_score(train_targets, train_predictions)
        val_r2 = r2_score(val_targets, val_predictions)

        epoch_time = time.time() - epoch_start_time

        history['train_loss'].append(train_loss)
        history['val_loss'].append(val_loss)
        history['train_mse'].append(train_mse)
        history['val_mse'].append(val_mse)
        history['train_rmse'].append(train_rmse)
        history['val_rmse'].append(val_rmse)
        history['learning_rate'].append(current_lr)
        history['epoch_time'].append(epoch_time)

    total_time = time.time() - start_time
    model.load_state_dict(best_model)

    # Calculate and save metrics
    train_metrics = calculate_metrics(model, train_loader, device)
    val_metrics = calculate_metrics(model, val_loader, device)
    test_metrics = calculate_metrics(model, test_loader, device)

    metrics = {
        'train_loss': train_metrics['loss'],
        'train_mse': train_metrics['mse'],
        'train_mae': train_metrics['mae'],
        'train_rmse': train_metrics['rmse'],
        'train_r2': train_metrics['r2'],
        'val_loss': val_metrics['loss'],
        'val_mse': val_metrics['mse'],
        'val_mae': val_metrics['mae'],
        'val_rmse': val_metrics['rmse'],
        'val_r2': val_metrics['r2'],
        'test_loss': test_metrics['loss'],
        'test_mse': test_metrics['mse'],
        'test_mae': test_metrics['mae'],
        'test_rmse': test_metrics['rmse'],
        'test_r2': test_metrics['r2'],
        'min_lr': min_lr,
        'max_lr': max_lr,
        'total_time': total_time,
        'average_epoch_time': sum(history['epoch_time']) / len(history['epoch_time'])
    }
    
    print(f"\nTraining completed in {total_time:.2f} seconds")
    print(f"Best model found at epoch {best_epoch}/{epochs}")
    print(f"Best Train Loss: {history['train_loss'][best_epoch-1]:.4f}, Best Val Loss: {history['val_loss'][best_epoch-1]:.4f}")
    print(f"Best Train MSE: {history['train_mse'][best_epoch-1]:.4f}, Best Val MSE: {history['val_mse'][best_epoch-1]:.4f}")
    print(f"Best Train RMSE: {history['train_rmse'][best_epoch-1]:.4f}, Best Val RMSE: {history['val_rmse'][best_epoch-1]:.4f}")

    # Save figures for this fold
    os.makedirs(f"models/Regression/{dataset_name}/ViT/{model_name}", exist_ok=True)
    plot_metric(history['train_loss'], history['val_loss'], 'Loss', dataset_name, model_name)
    plot_metric(history['train_mse'], history['val_mse'], 'MSE', dataset_name, model_name)
    plot_metric(history['train_rmse'], history['val_rmse'], 'RMSE', dataset_name, model_name)
    plot_learning_rate(history['learning_rate'], dataset_name, model_name)

    # Save metrics to a file
    os.makedirs(f'logs/Regression/{dataset_name}/ViT/{model_name}', exist_ok=True)
    with open(f'logs/Regression/{dataset_name}/ViT/{model_name}/metrics.txt', 'w') as f:
        for key, value in metrics.items():
            f.write(f'{key}: {value}\n')

    # Save best model
    model_save_path = f"models/Regression/{dataset_name}/ViT/{model_name}/best_model.pth"
    os.makedirs(os.path.dirname(model_save_path), exist_ok=True)
    torch.save(best_model, model_save_path)
    print(f"Best model saved to {model_save_path}")
            
    del model
    torch.cuda.empty_cache()
    gc.collect()

    return metrics

def calculate_metrics(model, data_loader, device):
    model.eval()
    loss_fn = nn.MSELoss()
    total_loss = 0
    all_targets = []
    all_predictions = []

    with torch.no_grad():
        for img_data, targets in data_loader:
            img_data, targets = img_data.to(device, non_blocking=True), targets.to(device, non_blocking=True)
            outputs = model(img_data)
            loss = loss_fn(outputs, targets)
            total_loss += loss.item()
            all_targets.append(targets.cpu().numpy())
            all_predictions.append(outputs.cpu().numpy())

    all_targets = np.concatenate(all_targets)
    all_predictions = np.concatenate(all_predictions)

    mse = mean_squared_error(all_targets, all_predictions, multioutput='raw_values')
    mae = mean_absolute_error(all_targets, all_predictions, multioutput='raw_values')
    rmse = np.sqrt(mse)
    r2 = r2_score(all_targets, all_predictions, multioutput='raw_values')

    metrics = {
        'loss': total_loss / len(data_loader.dataset),
        'mse': mse,
        'mae': mae,
        'rmse': rmse,
        'r2': r2
    }
    return metrics 


def safe_compile_and_fit(model, train_loader, val_loader, test_loader, 
                         dataset_name, model_name, batch_size=64, epochs=100, 
                         min_lr=1e-3, max_lr=1, device='cuda', weight_decay=1e-2):
    """A wrapper to safely compile and fit a model, with error handling."""
    try:
        if model is None:
            print(f"Model {model_name} is None")
            return None
        else:
            return compile_and_fit(
                model, train_loader, val_loader, test_loader,
                dataset_name, model_name, batch_size=batch_size, epochs=epochs,
                min_lr=min_lr, max_lr=max_lr, device=device, weight_decay=weight_decay
            )
    except Exception as e:
        print(f"Failed to compile and fit {model_name}: {str(e)}")
        return None
    finally:
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        gc.collect()


import traceback

def try_create_model(model_class, patch_size, imgs_shape):
    try:
        model = model_class(imgs_shape[1:], patch_size)
        
        # Test the model with a sample input
        sample_input = torch.randn(4, *imgs_shape)
        output = model(sample_input)
        
        print(f"Successfully created and tested {model_class.__name__}")
        
        return model
    except Exception as e:
        print(f"Error creating or testing {model_class.__name__}: {str(e)}")
        traceback.print_exc()
        return None


def plot_metric(train_metric, val_metric, metric_name, dataset_name, model_name):
    plt.figure()
    plt.plot(train_metric, label=f'Train {metric_name}')
    plt.plot(val_metric, label=f'Validation {metric_name}')
    plt.xlabel('Epoch')
    plt.ylabel(metric_name)
    plt.legend()
    plt.title(f'{metric_name} vs. Epoch')
    plt.savefig(f"models/Regression/{dataset_name}/ViT/{model_name}/{metric_name.lower()}_plot.png")
    plt.close()


def plot_learning_rate(learning_rates, dataset_name, model_name):
    plt.figure()
    plt.plot(learning_rates)
    plt.xlabel('Epoch')
    plt.ylabel('Learning Rate')
    plt.title('Learning Rate vs. Epoch')
    plt.savefig(f"models/Regression/{dataset_name}/ViT/{model_name}/learning_rate_plot.png")
    plt.close()


def find_best_model(base_path):
    best_rmse = float('inf')
    best_folder = None

    # Walk through all directories and files in the base path
    for root, dirs, files in os.walk(base_path):
        for file in files:
            if file == f'metrics.txt':
                file_path = os.path.join(root, file)
                
                # Read metrics from the file
                with open(file_path, 'r') as f:
                    metrics = f.read()
                
                # Parse the metrics into a dictionary
                metrics_dict = {}
                for line in metrics.splitlines():
                    key, value = line.split(': ')
                    metrics_dict[key.strip()] = float(value.strip())
                
                # Check if the current folder has a better validation loss
                if metrics_dict['test_rmse'] < best_rmse:
                    best_rmse = metrics_dict['test_rmse']
                    best_folder = root
    
    return best_folder, best_rmse

import os

def read_metrics(file_path):
    metrics = {}
    with open(file_path, 'r') as file:
        for line in file:
            key, value = line.split(': ')
            metrics[key.strip()] = float(value.strip())
    return metrics

def rename_folder(old_folder_path, prefix):
    folder_name = os.path.basename(old_folder_path)
    new_folder_name = f"{prefix}_{folder_name}"
    parent_dir = os.path.dirname(old_folder_path)
    new_folder_path = os.path.join(parent_dir, new_folder_name)
    os.rename(old_folder_path, new_folder_path)
    return new_folder_path

def process_folders(root_dir):
    prefixes = ["TINTO", "BarGraph", "Combination", "DistanceMatrix", "IGTD", "REFINED", "SuperTML", "FeatureWrap", "BIE"]
    best_folders = []

    for prefix in prefixes:
        matching_folders = [f for f in os.listdir(root_dir) if f.startswith(prefix) and os.path.isdir(os.path.join(root_dir, f))]
        if matching_folders:
            best_folder = None
            best_test_rmse = float('inf')
            for folder in matching_folders:
                metrics_file = os.path.join(root_dir, folder, 'metrics.txt')
                if os.path.exists(metrics_file):
                    metrics = read_metrics(metrics_file)
                    if metrics['test_rmse'] < best_test_rmse:
                        best_test_rmse = metrics['test_rmse']
                        best_folder = folder
            if best_folder:
                new_path = rename_folder(os.path.join(root_dir, best_folder), "TOP")
                best_folders.append(new_path)
    
    if best_folders:
        overall_best_folder = None
        overall_best_test_rmse = float('inf')
        for folder in best_folders:
            metrics_file = os.path.join(folder, 'metrics.txt')
            if os.path.exists(metrics_file):
                metrics = read_metrics(metrics_file)
                if metrics['test_rmse'] < overall_best_test_rmse:
                    overall_best_test_rmse = metrics['test_rmse']
                    overall_best_folder = folder
        if overall_best_folder:
            rename_folder(overall_best_folder, "BEST")
        
    return best_folders

# ------------------------- MAIN FUNCTION ------------------------- #
def main():
    # --------------------- SET YOUR VARIABLES HERE ---------------------- #
    SEED = 64
    dataset_name = 'boston'      # Choose your dataset name
    batch_size = 32
    problem_type = "regression"  # [‘supervised’, ‘unsupervised’, ‘regression’]
    epochs = 100                 # Default training epochs

    # Path for dataset and logs
    results_path = f'logs/Regression/{dataset_name}/ViT_Regression'
    
    # ---------------------- SET RANDOM SEEDS ---------------------- #
    torch.manual_seed(SEED)
    random.seed(SEED)
    np.random.seed(SEED)

    # ---------------------- LOAD DATA ---------------------- #
    df = pd.read_csv(f"../Dataset/Regression/{dataset_name}.csv")
    print("Dataset shape:", df.shape)
    print(df.head())

    # ---------------------- IMAGE MODEL SELECTION ---------------------- #
    # You can switch out TINTO for REFINED, IGTD, BarGraph, DistanceMatrix, etc.
    #Select the model and the parameters
    problem_type = "regression"
    image_model = TINTO(problem= problem_type, blur=True, random_seed=SEED)
    #image_model = REFINED(problem= problem_type,hcIterations=5)
    #image_model = IGTD(problem= problem_type)
    #image_model = BarGraph(problem= problem_type)
    #image_model = DistanceMatrix(problem= problem_type)
    #image_model = Combination(problem= problem_type)
    #image_model = SuperTML(problem= problem_type)
    #image_model = FeatureWrap(problem= problem_type)
    #image_model = BIE(problem= problem_type)
    name = f"TINTO_blur"

    #Define the dataset path and the folder where the images will be saved
    images_folder = f"./Synthetic_images/Regression/{dataset_name}/images_{dataset_name}_{name}"

    # ---------------------- PREPARE DATA LOADERS ---------------------- #
    train_loader, val_loader, test_loader, attributes, imgs_shape = load_and_preprocess_data(
        df, images_folder, image_model, problem_type, batch_size=batch_size, seed=SEED
    )

    # ---------------------- TRAIN & EVALUATE MODEL1 ---------------------- #
    patch_size = find_divisors(imgs_shape[1])
    print("Patch size candidates:", patch_size)

    patch_size = 4 # Change as needed

    model1 = try_create_model(Model1, patch_size, imgs_shape)
    model1_metrics = safe_compile_and_fit(
        model1, train_loader, val_loader, test_loader, 
        dataset_name, f"{name}_Model1_patch{patch_size}", 
        min_lr=1e-4, max_lr=1e-2, epochs=epochs
    )

    # ---------------------- TRAIN & EVALUATE MODEL2 ---------------------- #
    model2 = try_create_model(Model2, patch_size, imgs_shape)
    model2_metrics = safe_compile_and_fit(
        model2, train_loader, val_loader, test_loader, 
        dataset_name, f"{name}_Model2_patch{patch_size}", 
        min_lr=1e-4, max_lr=1e-2, epochs=epochs
    )

    # ---------------------- PRINT RESULTS ---------------------- #
    if model1_metrics:
        print("Model 1 Metrics:", model1_metrics)
    if model2_metrics:
        print("Model 2 Metrics:", model2_metrics)

    # ---------------------- FIND AND ANNOTATE BEST MODELS ---------------------- #
    base_path = f"logs/Regression/{dataset_name}/ViT/"
    best_folders = process_folders(base_path)
    print(f"Best folders marked: {best_folders}")


# Standard Python entry point
if __name__ == "__main__":
    main()



# Future imports
from __future__ import division

# Standard library imports
import math
import os

# Third-party library imports
import numpy as np
import pandas as pd
from PIL import Image, ImageDraw, ImageFont
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

# Typing imports
from typing import Union

# Local application/library imports
from TINTOlib.abstractImageMethod import AbstractImageMethod

###########################################################
################    SuperTML    ###########################
###########################################################

class SuperTML(AbstractImageMethod):
    """
    SuperTML: A method for transforming tabular data into synthetic images inspired by natural language processing techniques.

    This method draws data values directly onto a 1-channel image where each feature is assigned a region and represented in text.
    It supports two approaches:
    - SuperTML_EF: Each feature has an equal-sized region and font size.
    - SuperTML_VF: Each feature has a region and font size proportional to its relevance, with relevance calculated using Random Forest feature importance.

    Parameters:
    ----------
    problem : str, optional
        The type of problem, defining how the images are grouped. 
        Default is 'supervised'. Valid values: ['supervised', 'unsupervised', 'regression'].
    normalize : bool, optional
        If True, normalizes input data using MinMaxScaler. 
        Default is True. Valid values: [True, False].
    verbose : bool, optional
        Show execution details in the terminal. 
        Default is False. Valid values: [True, False].
    image_pixels : int, optional
        The number of pixels used to create the image (one side). 
        Total pixels = pixels * pixels. Default is 224. Valid values: integer.
    feature_importance : bool, optional
        If False, SuperTML-EF is used (Equal Font), where all features are displayed with equal font sizes. 
        If True, SuperTML-VF is used (Variable Font), where the font size of each feature is proportional to its importance. 
        Default is False. Valid values: [True, False].
    font_size : int, optional
        The size of the font used to render text on the generated images. 
        Default is 10. Valid values: integer.
    random_seed : int, optional
        Seed for reproducibility. The method uses random forest for feature importance.
        Default is 1. Valid values: integer.
    """
    ###### default values ###############
    default_pixels = 224  # Default image size (width and height in pixels)

    default_feature_importance = False  # Use feature importance (False = SuperTML-EF, True = SuperTML-VF)
    default_font_size = 10  # Default font size

    default_random_seed = 1  # Default random seed for reproducibility
    
    def __init__(
        self,
        problem=None,
        normalize=None,
        verbose=None,
        pixels=default_pixels,
        feature_importance=default_feature_importance,
        font_size=default_font_size,
        random_seed=default_random_seed,
    ):
        super().__init__(problem=problem, verbose=verbose, normalize=normalize)

        self.image_pixels = pixels

        self.feature_importance = feature_importance
        self.font_size = font_size

        self.random_seed = random_seed

    def __saveSupervised(self, y, i, image):
        extension = 'png'  # eps o pdf
        subfolder = str(int(y)).zfill(2)  # subfolder for grouping the results of each class
        name_image = str(i).zfill(6)
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image + '.' + extension)
        # Subfolder check
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        image.save(route_complete)
        route_relative = os.path.join(subfolder, name_image+ '.' + extension)
        return route_relative
    
    def __saveRegressionOrUnsupervised(self, i, image):
        extension = 'png'  # eps o pdf
        subfolder = "images"
        name_image = str(i).zfill(6) + '.' + extension
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image)
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")
        image.save(route_complete)

        route_relative = os.path.join(subfolder, name_image)
        return route_relative
    
    # def calculate_feature_importance(self, data):
    #     # Dummy implementation, replace with actual feature importance calculation
    #     # Example: {'feature_0': 0.1, 'feature_1': 0.4, 'feature_2': 0.5}
    #     return {f'feature_{i}': np.random.rand() for i in range(data.shape[1])}

    def check_overlap(self, x, y, text_width, text_height, positions):
        for px, py, pw, ph in positions:
            if not (x + text_width < px or x > px + pw or y + text_height < py or y > py + ph):
                return True
        return False

    def __event2img(self,event: np.ndarray):
        # SuperTML-VF
        if self.feature_importance:
            padding = 5     # Padding around the texts

            feature_importances = self.feature_importances
            max_feature_importances = max(feature_importances.tolist())

            img = Image.fromarray(np.zeros([self.image_pixels, self.image_pixels, 3]), 'RGB')
            draw = ImageDraw.Draw(img)

            sorted_features = sorted(zip(event, feature_importances), key=lambda x: x[1], reverse=True)
            positions = []

            for i,(feature,importance) in enumerate(sorted_features):
                # The font size of this feature is relative to the ratio of this importante vs the most important feature
                ratio = (importance / max_feature_importances)
                font_size = max(int(self.font_size * ratio), 1)
                font = ImageFont.truetype("arial.ttf", font_size)

                text = f'{feature:.3f}'

                bbox = draw.textbbox((0, 0), text, font=font)
                text_width = bbox[2] - bbox[0]
                text_height = bbox[3] - bbox[1]

                placed = False
                for y in range(0, self.image_pixels - text_height, 1):
                    if placed:
                        break
                    for x in range(0, self.image_pixels - text_width, 1):
                        if not self.check_overlap(x, y, text_width, text_height, positions):
                            positions.append((x, y, text_width+padding, text_height+padding))
                            draw.text((x, y), text, fill=(255, 255, 255), font=font)
                            placed = True
                            break

            return img

        # SuperTML-EF
        else:
            cell_width = self.image_pixels // self.columns
            rows = math.ceil(len(event) / self.columns)
            cell_height = self.image_pixels // rows

            font = ImageFont.truetype("arial.ttf", self.font_size)
            img = Image.fromarray(np.zeros([self.image_pixels, self.image_pixels, 3]), 'RGB')
            draw = ImageDraw.Draw(img)

            for i, f in enumerate(event):
                x = ((i % self.columns)) * cell_width
                y = (i // self.columns) * cell_height

                text = f'{f:.3f}'
                
                bbox = draw.textbbox((0, 0), text, font=font)
                text_width = bbox[2] - bbox[0]
                text_height = bbox[3] - bbox[1]

                text_x = x + (cell_width - text_width) / 2
                text_y = y + (cell_height - text_height) / 2

                draw.text(
                    (text_x, text_y),
                    text,
                    fill=(255, 255, 255),
                    font=font,
                )

            return img
        
    def calculate_feature_importances(self, X: np.ndarray, y: np.ndarray) -> None:
        """
        Calculates feature importances using a Random Forest model.

        Arguments
        ---------
        X: np.ndarray
            Feature array
        y: np.ndarray
            Targets array
        """

        max_selection = 100_000

        indices = np.arange(X.shape[0])
        np.random.shuffle(indices)

        # Split the data into training and testing sets
        if self.problem == 'supervised':
            model = RandomForestClassifier(random_state=self.random_seed, n_jobs=-1)
        else:
            model = RandomForestRegressor(random_state=self.random_seed, n_jobs=-1)

        # Fit the model
        model.fit(X[indices][:max_selection], y[indices][:max_selection])

        # Update the feature importances
        self.feature_importances = model.feature_importances_

    def _fitAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        X = x.values
        Y = y.values if y is not None else None

        if self.feature_importance:
            # Calculate the feature importance for SuperTML-VF
            self.calculate_feature_importances(X, Y)
        else:
            # Calculate the number of columns
            self.columns = math.ceil(math.sqrt(X.shape[1]))

    def _transformAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        X = x.values
        Y = y.values if y is not None else None

        # Variable for regression problem
        imagesRoutesArr = []

        try:
            os.makedirs(self.folder)
            if self.verbose:
                print("The folder was created " + self.folder + "...")
        except:
            if self.verbose:
                print("The folder " + self.folder + " is already created...")
        for i in range(X.shape[0]):

            image = self.__event2img(X[i])

            if self.problem == "supervised":
                route = self.__saveSupervised(Y[i], i, image)
                imagesRoutesArr.append(route)
            elif self.problem == "unsupervised" or self.problem == "regression":
                route = self.__saveRegressionOrUnsupervised(i, image)
                imagesRoutesArr.append(route)
            else:
                print("Wrong problem definition. Please use 'supervised', 'unsupervised' or 'regression'")
        
        if self.problem == "supervised" :
            data={'images':imagesRoutesArr,'class':Y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/supervised.csv", index=False)
        elif self.problem == "unsupervised":
            data = {'images': imagesRoutesArr}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/unsupervised.csv", index=False)
        elif self.problem == "regression":
            data = {'images': imagesRoutesArr,'values':Y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/regression.csv", index=False)     



# Standard library imports
import os
import shutil

# Third-party library imports
import bitstring
import matplotlib
import matplotlib.image
import numpy as np
import pandas as pd
from PIL import Image

# Typing imports
from typing import Iterator, List, Union

# Local application/library imports
from TINTOlib.abstractImageMethod import AbstractImageMethod

###########################################################
################    Combination    ########################
###########################################################

class Combination(AbstractImageMethod):
    """
    Combination: Combines the Distance Matrix and BarGraph representations.

    This method generates a hybrid visualization by combining the outputs of the 
    Distance Matrix and BarGraph methods into a single image representation.

    Parameters:
    ----------
    problem : str, optional
        The type of problem, defining how the images are grouped. 
        Default is 'supervised'. Valid values: ['supervised', 'unsupervised', 'regression'].
    normalize : bool, optional
        If True, normalizes input data using MinMaxScaler. 
        Default is True. Valid values: [True, False].
    verbose : bool, optional
        Show execution details in the terminal. 
        Default is False. Valid values: [True, False].
    zoom : int, optional
        Multiplication factor determining the size of the saved image relative to the original size. 
        Default is 1. Valid values: integer > 0.
    """
    ###### Default values ###############
    default_zoom = 1  # Rescale factor for saving the image

    def __init__(
        self,
        problem = None,
        normalize=None,
        verbose = None,
        zoom=default_zoom,
    ):
        super().__init__(problem=problem, verbose=verbose, normalize=normalize)

        self.zoom = zoom

    def __saveSupervised(self, y, i, image):
        extension = 'png'  # eps o pdf
        subfolder = str(int(y)).zfill(2)  # subfolder for grouping the results of each class
        name_image = str(i).zfill(6)
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image + '.' + extension)
        # Subfolder check
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        img = Image.fromarray(np.uint8(np.squeeze(image) * 255))
        img.save(route_complete)

        route_relative = os.path.join(subfolder, name_image+ '.' + extension)
        return route_relative

    def __saveRegressionOrUnsupervised(self, i, image):
        extension = 'png'  # eps o pdf
        subfolder = "images"
        name_image = str(i).zfill(6) + '.' + extension
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image)
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")
        img = Image.fromarray(np.uint8(np.squeeze(image) * 255))
        img.save(route_complete)

        route_relative = os.path.join(subfolder, name_image)
        return route_relative
    
    def _fitAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        """
        Fit method for stateless transformers. Does nothing and returns self.
        """
        return self
    
    def _transformAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        X = x.values
        Y = y.values if y is not None else None

        imagesRoutesArr = []

        n_columns = X.shape[1]
        # Matrix for level 1 calculations
        imgI = np.empty((n_columns, n_columns))

        pixel_width, gap = 1, 0
        assert (pixel_width*n_columns + (n_columns+1)*gap) == n_columns
        top_padding, bottom_padding = pixel_width, pixel_width
        max_bar_height = n_columns - (bottom_padding + top_padding)
        step_column = gap + pixel_width

        # Matrix for level 2 calculations
        imgage = np.zeros([n_columns, n_columns, 1])

        # Matrix for level 3 calculations
        lvl3img = np.empty((n_columns, n_columns))

        # for each instance
        for ins,dataInstance in enumerate(X):
            """LEVEL - 1 (MATRIX)"""
            for i in range(n_columns):
                for j in range(n_columns):
                    imgI[i][j] = dataInstance[i] - dataInstance[j]
            # Normalize matrix
            image_norm = (imgI - np.min(imgI)) / (np.max(imgI) - np.min(imgI))
            # Apply zoom
            imgI1 = np.repeat(np.repeat(image_norm, self.zoom, axis=0), self.zoom, axis=1)

            """LEVEL - 2 (BARS)"""
            # TODO: reorder the columns
            imgage = np.zeros([n_columns, n_columns, 1])
            bar_heights = np.floor(X[ins] * max_bar_height).astype(np.int64)
            for i_bar,val_bar in enumerate(bar_heights):
                imgage[
                    top_padding : (top_padding + val_bar),                              # The height of the column
                    (gap+(step_column*i_bar)) : (gap+(step_column*i_bar)) + pixel_width # The width of the column
                ] = 1
            # Apply zoom
            imgI2 = np.repeat(np.repeat(imgage, self.zoom, axis=0), self.zoom, axis=1)

            """LEVEL - 3"""
            # Fill the matrix
            lvl3img[:, :] = dataInstance[:, np.newaxis]
            # Normalize the matrix
            lvl3 = (lvl3img - np.min(lvl3img)) / (np.max(lvl3img) - np.min(lvl3img))
            # Apply zoom
            imgI3 = np.repeat(np.repeat(lvl3, self.zoom, axis=0), self.zoom, axis=1)

            # Combine ALL img
            imgFinal = np.dstack((imgI1,imgI2,imgI3))

            if self.problem == "supervised":
                route = self.__saveSupervised(Y[ins], ins, imgFinal)
                imagesRoutesArr.append(route)
            elif self.problem == "unsupervised" or self.problem == "regression":
                route = self.__saveRegressionOrUnsupervised(ins, imgFinal)
                imagesRoutesArr.append(route)
            else:
                print("Wrong problem definition. Please use 'supervised', 'unsupervised' or 'regression'")

        if self.problem == "supervised":
            data = {'images': imagesRoutesArr, 'class': Y}
            supervisedCSV = pd.DataFrame(data=data)
            supervisedCSV.to_csv(self.folder + "/supervised.csv", index=False)
        elif self.problem == "unsupervised":
            data = {'images': imagesRoutesArr}
            unsupervisedCSV = pd.DataFrame(data=data)
            unsupervisedCSV.to_csv(self.folder + "/unsupervised.csv", index=False)
        elif self.problem == "regression":
            data = {'images': imagesRoutesArr, 'values': Y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/regression.csv", index=False)



# Standard library imports
import gc
import math
import os

# Third-party library imports
import matplotlib
import matplotlib.image
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

# Typing imports
from typing import Union

# Local application/library imports
from TINTOlib.abstractImageMethod import AbstractImageMethod

###########################################################
################    TINTO    ##############################
###########################################################


class TINTO(AbstractImageMethod):
    """
    TINTO: A class for transforming tabular data into synthetic images by projecting it into a two-dimensional space 
    and applying visualization techniques.
    
    Parameters:
    ----------
    problem : str, optional
        The type of problem, defining how the images are grouped. 
        Default is 'supervised'. Valid values: ['supervised', 'unsupervised', 'regression'].
    normalize : bool, optional
        If True, normalizes input data using MinMaxScaler. 
        Default is True. Valid values: [True, False].
    verbose : bool, optional
        Show execution details in the terminal. 
        Default is False. Valid values: [True, False].
    pixels : int, optional
        The number of pixels used to create the image (one side). 
        Total pixels = pixels * pixels. Default is 20. Valid values: integer.
    algorithm : str, optional
        Select the dimensionality reduction algorithm. 
        Default is 'PCA'. Valid values: ['PCA', 't-SNE'].
    blur : bool, optional
        Activate or deactivate the blurring option. 
        Default is False. Valid values: [True, False].
    submatrix : bool, optional
        Specifies whether to use a submatrix for blurring. 
        Default is True. Valid values: [True, False].
    amplification : float, optional
        Only used when `blur=True`. Specifies the blurring amplification. 
        Default is `np.pi`. Valid values: float.
    distance : int, optional
        Only used when `blur=True`. Specifies the blurring distance (number of pixels). 
        Default is 2. Valid values: integer.
    steps : int, optional
        Only used when `blur=True`. Specifies the number of blurring steps. 
        Default is 4. Valid values: integer.
    option : str, optional
        Only used when `blur=True`. Technique for handling overlapping pixels. 
        Default is 'mean'. Valid values: ['mean', 'maximum'].
    times : int, optional
        Only used when `algorithm='t-SNE'`. Specifies the replication times in t-SNE. 
        Default is 4. Valid values: integer.
    zoom : int, optional
        Multiplication factor determining the size of the saved image relative to the original size. 
        Values greater than 1 increase the image size proportionally. 
        Default is 1. Valid values: integer.
    random_seed : int, optional
        Seed for reproducibility. 
        Default is 1. Valid values: integer.

    Attributes:
    ----------
    error_pos : bool
        Indicates overlap of characteristic pixels during image creation.
    """
    ###### default values ###############
    default_pixels = 20  # Image's Pixels (one side)

    default_algorithm = "PCA"  # Dimensionality reduction algorithm (PCA or t-SNE)
    default_blur = False  # Activate blurring option

    default_submatrix = True  # Use or not use submatrix
    default_amplification = np.pi  # Amplification factor in blurring
    default_distance = 2  # Distance in blurring (number of pixels)
    default_steps = 4  # Steps in blurring
    default_option = 'mean'  # Option in blurring ('mean' or 'maximum')

    default_times = 4  # Replication times in t-SNE
    default_train_m = True  # Use training matrix

    default_random_seed = 1  # Seed for reproducibility
    default_zoom = 1  # Zoom level

    def __init__(
        self,
        problem=None,
        normalize=None,
        verbose=None,
        pixels=default_pixels,
        algorithm=default_algorithm,
        blur=default_blur,
        submatrix=default_submatrix,
        amplification=default_amplification,
        distance=default_distance,
        steps=default_steps,
        option=default_option,
        times=default_times,
        train_m=default_train_m,
        zoom=default_zoom,
        random_seed=default_random_seed,
    ):
        super().__init__(problem=problem, verbose=verbose, normalize=normalize)

        self.pixels = pixels

        self.algorithm = algorithm
        self.blur = blur
        
        self.submatrix = submatrix
        self.amplification = amplification
        self.distance = distance
        self.steps = steps
        self.option = option

        self.times = times
        self.train_m = train_m

        self.zoom = zoom
        self.random_seed = random_seed

        self.error_pos = False  # Indicates the overlap of characteristic pixels.

    def __square(self, coord):
        """
        This function has the purpose of being able to create the square delimitation of the resulting image.
        The steps it performs in the order of the code are as follows:
            - Calculate the average of the points $(x,y)$.
            - Centres the data at the point $(0,0)$.
            - Calculate the distance from the centre to the points.
            - The larger distance of \texttt{dista}, transforms it to integer.
            - Calculate the vertices of the square.
            - Move the points to quadrant $4$.
            - Transfers the vertices as well.
            - Returns the values, coordinates, and vertices.
        """
        m = np.mean(coord, axis=0).reshape((1, 2))
        coord_new = coord - m
        dista = (coord_new[:, 0] ** 2 + coord_new[:, 1] ** 2) ** 0.5
        maxi = math.ceil(max(dista))
        vertices = np.array([[-maxi, maxi], [-maxi, -maxi], [maxi, -maxi], [maxi, maxi]])
        coord_new = coord_new - vertices[0]
        vertices = vertices - vertices[0]
        return coord_new, vertices

    def __m_imagen(self, coord, vertices):
        """
        This function obtain the coordinates of the matrix. This function has
        the following specifications:
            - Create a matrix of coordinates and vertices.
            - Transform the coordinates into indices for the matrix.
            - Fill in the positions of the features.
            - Finally, a conditional is created if the features were grouped
              in the same position.
        """
        pixels = self.pixels
        size = (pixels, pixels)
        matrix = np.zeros(size)

        coord_m = (coord / vertices[2, 0]) * (pixels - 1)
        coord_m = np.round(abs(coord_m))

        for i, j in zip(coord_m[:, 1], coord_m[:, 0]):
            matrix[int(i), int(j)] = 1

        if np.count_nonzero(matrix != 0) != coord.shape[0]:
            return coord_m, matrix, True
        else:
            return coord_m, matrix, False

    def __createFilter(self):
        """
        In this function a filter is created since a matrix of size "2*distance*total_steps+1"
        is being created to act as a "filter", which covers the whole circular space of the minutiae
        determined by the distance and by the total number of steps.
        This "filter", which is a matrix, would be multiplied with a scalar, which is the intensity value.
        Finally, this resulting matrix is placed as a submatrix within the final matrix where the centre
        of the submatrix would be the position of the characteristic pixel.
        """
        distance = self.distance
        steps = self.steps
        amplification = self.amplification
        size_filter = int(2 * distance * steps + 1)
        center_x = distance * steps
        center_y = distance * steps
        filter = np.zeros([size_filter, size_filter])

        for step in reversed(range(steps)):
            r_actual = int(distance * (step + 1))  # current radius from largest to smallest

            # Function of intensity
            intensity = min(amplification * 1 / (np.pi * r_actual ** 2), 1)

            # Delimitation of the area
            lim_inf_i = max(center_x - r_actual - 1, 0)
            lim_sup_i = min(center_x + r_actual + 1, size_filter)
            lim_inf_j = max(center_y - r_actual - 1, 0)
            lim_sup_j = min(center_y + r_actual + 1, size_filter)

            # Allocation of values
            for i in range(lim_inf_i, lim_sup_i):
                for j in range(lim_inf_j, lim_sup_j):
                    if (center_x - i) ** 2 + (center_y - j) ** 2 <= r_actual ** 2:
                        filter[i, j] = intensity
        filter[center_x, center_y] = 1
        return filter

    def __blurringFilterSubmatrix(self, matrix, filter, values, coordinates):
        """
       This function is to be able to add more ordered contextual information to the image through the
       classical painting technique called blurring. This function develops the following main steps:
       - Take the coordinate matrix of the characteristic pixels.
       - Create the blurring according to the number of steps taken in a loop with the
         following specifications:
            - Delimit the blurring area according to $(x,y)$ on an upper and lower boundary.
            - Set the new intensity values in the matrix, taking into account that if there is
              pixel overlap, the maximum or average will be taken as specified.
        """
        option = self.option

        iter_values = iter(values)
        size_matrix = matrix.shape[0]
        size_filter = filter.shape[0]
        matrix_extended = np.zeros([size_filter + size_matrix, size_filter + size_matrix])
        matrix_add = np.zeros([size_filter + size_matrix, size_filter + size_matrix])
        center_filter = int((size_filter - 1) / 2)
        for i, j in coordinates:
            i = int(i)
            j = int(j)
            value = next(iter_values)
            submatrix = filter * value

            # Delimitación del área
            lim_inf_i = i
            lim_sup_i = i + 2 * center_filter + 1
            lim_inf_j = j
            lim_sup_j = j + 2 * center_filter + 1

            if option == 'mean':
                matrix_extended[lim_inf_i:lim_sup_i, lim_inf_j:lim_sup_j] += submatrix
                matrix_add[lim_inf_i:lim_sup_i, lim_inf_j:lim_sup_j] += (submatrix > 0) * 1
            elif option == 'maximum':
                matrix_extended[lim_inf_i:lim_sup_i, lim_inf_j:lim_sup_j] = np.maximum(
                    matrix_extended[lim_inf_i:lim_sup_i, lim_inf_j:lim_sup_j], submatrix)

        if option == 'mean':
            matrix_add[matrix_add == 0] = 1
            matrix_extended = matrix_extended / matrix_add

        matrix_final = matrix_extended[center_filter:-center_filter - 1, center_filter:-center_filter - 1]

        return matrix_final

    def __blurringFilter(self, matrix, coordinates):
        """
       This function is to be able to add more ordered contextual information to the image through the
       classical painting technique called blurring. This function develops the following main steps:
       - Take the coordinate matrix of the characteristic pixels.
       - Create the blurring according to the number of steps taken in a loop with the
         following specifications:
            - Delimit the blurring area according to $(x,y)$ on an upper and lower boundary.
            - Set the new intensity values in the matrix, taking into account that if there is
              pixel overlap, the maximum or average will be taken as specified.
        """

        x = int(coordinates[1])
        y = int(coordinates[0])
        valor_central = matrix[x, y]

        for p in range(self.steps):
            r_actual = int(matrix.shape[0] * self.distance * (p + 1))  # radio actual  de mayor a menor

            # Funcion de intensidad
            intensidad = min(self.amplification * valor_central / (np.pi * r_actual ** 2), valor_central)

            # Delimitación del área
            lim_inf_i = max(x - r_actual - 1, 0)
            lim_sup_i = min(x + r_actual + 1, matrix.shape[0])
            lim_inf_j = max(y - r_actual - 1, 0)
            lim_sup_j = min(y + r_actual + 1, matrix.shape[1])

            # Colocación de valores
            for i in range(lim_inf_i, lim_sup_i):
                for j in range(lim_inf_j, lim_sup_j):
                    if ((x - i) ** 2 + (y - j) ** 2 <= r_actual ** 2):
                        if (matrix[i, j] == 0):
                            matrix[i, j] = intensidad
                        elif (x != i and y != j):  # Sobreposición
                            if (self.option == 'mean'):
                                matrix[i, j] = (matrix[i, j] + intensidad) / 2
                            elif (self.option == 'maximum'):
                                matrix[i, j] = max(matrix[i, j], intensidad)
        return matrix


    def __saveSupervised(self, y, i, matrix_a):
        extension = 'png'  # eps o pdf
        subfolder = str(int(y)).zfill(2)  # subfolder for grouping the results of each class
        name_image = str(i).zfill(6)
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image + '.' + extension)
        # Subfolder check
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        # Repeat matrix to apply zoom
        matrix_a = np.repeat(np.repeat(matrix_a, self.zoom, axis=0), self.zoom, axis=1)
        matplotlib.image.imsave(route_complete, matrix_a, cmap='binary', format=extension, dpi=self.zoom)

        route_relative = os.path.join(subfolder, name_image+ '.' + extension)
        return route_relative

    def __saveRegressionOrUnsupervised(self, i, matrix_a):
        extension = 'png'  # eps o pdf
        subfolder = "images"
        name_image = str(i).zfill(6) + '.' + extension
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image)
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        # Repeat matrix to apply zoom
        matrix_a = np.repeat(np.repeat(matrix_a, self.zoom, axis=0), self.zoom, axis=1)
        matplotlib.image.imsave(route_complete, matrix_a, cmap='binary', format=extension, dpi=self.zoom)

        route_relative = os.path.join(subfolder, name_image)
        return route_relative

    def __imageSampleFilterSubmatrix(self, X, Y, coord, matrix):
        """
        This function creates the samples, i.e., the images. This function has the following specifications:
        - The first conditional performs the pre-processing of the images by creating the matrices.
        - Then the for loop generates the images for each sample. Some assumptions have to be taken into
          account in this step:
            - The samples will be created according to the number of targets. Therefore, each folder that is
              created will contain the images created for each target.
            - In the code, the images are exported in PNG format; this can be changed to any other format.
        """
        # Hyperparams
        amplification = self.amplification
        distance = self.distance
        steps = self.steps
        option = self.option
        imagesRoutesArr=[]

        # Generate the filter
        if distance * steps * amplification != 0:  # The function is only called if there are no zeros (blurring).
            filter = self.__createFilter()

        total = X.shape[0]
        # In this part, images are generated for each sample.
        for i in range(total):
            matrix_a = np.zeros(matrix.shape)
            if distance * steps * amplification != 0:  # The function is only called if there are no zeros (blurring).
                matrix_a = self.__blurringFilterSubmatrix(matrix_a, filter, X[i], coord)
            else:  # (no blurring)
                iter_values_X = iter(X[i])
                for eje_x, eje_y in coord:
                    matrix_a[int(eje_x), int(eje_y)] = next(iter_values_X)

            if self.problem == "supervised":
                route=self.__saveSupervised(Y[i],i,matrix_a)
                imagesRoutesArr.append(route)
            elif self.problem == "unsupervised" or self.problem == "regression":
                route = self.__saveRegressionOrUnsupervised( i, matrix_a)
                imagesRoutesArr.append(route)
            else:
                print("Wrong problem definition. Please use 'supervised', 'unsupervised' or 'regression'")

            #Verbose
            if self.verbose:
                print("Created ", str(i + 1), "/", int(total))

        if self.problem == "supervised":
            data = {'images': imagesRoutesArr, 'class': Y}
            supervisedCSV = pd.DataFrame(data=data)
            supervisedCSV.to_csv(self.folder + "/supervised.csv", index=False)
        elif self.problem == "unsupervised":
            data = {'images': imagesRoutesArr}
            unsupervisedCSV = pd.DataFrame(data=data)
            unsupervisedCSV.to_csv(self.folder + "/unsupervised.csv", index=False)
        elif self.problem == "regression":
            data = {'images': imagesRoutesArr, 'values': Y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/regression.csv", index=False)

        return matrix

    def __imageSampleFilter(self, X, Y, coord, matrix):
        """
        This function creates the samples, i.e., the images. This function has the following specifications:
        - The first conditional performs the pre-processing of the images by creating the matrices.
        - Then the for loop generates the images for each sample. Some assumptions have to be taken into
          account in this step:
            - The samples will be created according to the number of targets. Therefore, each folder that is
              created will contain the images created for each target.
            - In the code, the images are exported in PNG format; this can be changed to any other format.
        """
        imagesRoutesArr = []
        total = X.shape[0]
        # Aquí es donde se realiza el preprocesamiento siendo 'matriz' = 'm'
        if self.train_m:

            matrix_a = np.zeros(matrix.shape)

            for i in range(X.shape[0]):
                for j in range(X.shape[1]):
                    matrix_a[int(coord[j, 1]), int(coord[j, 0])] = X[i, j]
                    matrix_a = self.__blurringFilter(matrix_a, coord[j] )

            matriz = np.copy(matrix_a)
        if self.verbose:
            print("Generating images......")
        # En esta parte se generan las imágenes por cada muestra
        for i in range(total):
            matrix_a = np.copy(matriz)

            for j in range(X.shape[1]):
                matrix_a[int(coord[j, 1]), int(coord[j, 0])] = X[i, j]
                matrix_a = self.__blurringFilter(matrix_a, coord[j])

            for j in range(X.shape[1]):
                matrix_a[int(coord[j, 1]), int(coord[j, 0])] = X[i, j]

            if self.problem == "supervised":
                    route=self.__saveSupervised(Y[i],i,matrix)
                    imagesRoutesArr.append(route)
            elif self.problem == "unsupervised" or self.problem == "regression":
                    route = self.__saveRegressionOrUnsupervised( i, matrix_a)
                    imagesRoutesArr.append(route)
            else:
                    print("Wrong problem definition. Please use 'supervised', 'unsupervised' or 'regression'")

            #Verbose
            if self.verbose:
                print("Created ", str(i + 1), "/", int(total))

        if self.problem == "supervised":
            data = {'images': imagesRoutesArr, 'class': Y}
            supervisedCSV = pd.DataFrame(data=data)
            supervisedCSV.to_csv(self.folder + "/supervised.csv", index=False)
        elif self.problem == "unsupervised":
            data = {'images': imagesRoutesArr}
            unsupervisedCSV = pd.DataFrame(data=data)
            unsupervisedCSV.to_csv(self.folder + "/unsupervised.csv", index=False)
        elif self.problem == "regression":
            data = {'images': imagesRoutesArr, 'values': Y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/regression.csv", index=False)

        return matrix

    def __obtainCoord(self, X):
        """
        This function uses the dimensionality reduction algorithm in order to represent the characteristic
        pixels in the image. The specifications of this function are:
        - Perform a normalisation of (0,1) to be able to represent the pixels inside the square.
        - Transpose the matrix.
        - Set the dimensionality reduction algorithm, PCA or t-SNE.
        """

        labels = np.arange(X.shape[1])
        
        X_trans = X.T

        if self.verbose:
            print("Selected algorithm: " + self.algorithm)

        if self.algorithm == 'PCA':
            X_embedded = PCA(n_components=2, random_state=self.random_seed).fit(X_trans).transform(X_trans)
        elif self.algorithm == 't-SNE':
            for i in range(self.times):
                X_trans = np.append(X_trans, X_trans, axis=0)
                labels = np.append(labels, labels, axis=0)
            X_embedded = TSNE(n_components=2, random_state=self.random_seed, perplexity=50).fit_transform(X_trans)
        else:
            print("Error: Incorrect algorithm")
            X_embedded = np.random.rand(X.shape[1], 2)

        data_coord = {'x': X_embedded[:, 0], 'y': X_embedded[:, 1], 'Label': labels}
        dc = pd.DataFrame(data=data_coord)
        self.obtain_coord = dc.groupby('Label').mean().values

        del X_trans
        gc.collect()

    def __areaDelimitation(self):
        """
        This function performs the delimitation of the area
        """
        self.initial_coordinates, self.vertices = self.__square(self.obtain_coord)

    def __matrixPositions(self):
        """
        This function gets the positions in the matrix
        """
        self.pos_pixel_caract, self.m, self.error_pos = self.__m_imagen(self.initial_coordinates, self.vertices)

    def __createImage(self, X, Y, folder='prueba/'):
        """
        This function creates the images.
        """
        X = X.values
        Y = np.array(Y)
        try:
            os.makedirs(folder)
            if self.verbose:
                print("The folder was created " + folder + "...")
        except:
            if self.verbose:
                print("The folder " + folder + " is already created...")

        if self.submatrix:
            self.m = self.__imageSampleFilterSubmatrix(X, Y, self.pos_pixel_caract, self.m)
        else:
            self.m = self.__imageSampleFilter(X, Y, self.pos_pixel_caract, self.m)

    def _fitAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        if not self.blur:
            self.amplification = 0
            self.distance = 2
            self.steps = 0

        X = x.values
        Y = y.values if y is not None else None

        self.__obtainCoord(X)
        self.__areaDelimitation()
        self.__matrixPositions()

    def _transformAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        if not self.blur:
            self.amplification = 0
            self.distance = 2
            self.steps = 0

        if (y is None):
            y = np.zeros(x.shape[0])
        self.__createImage(x, y, self.folder)


# Standard library imports
import math
import os
import pickle
import platform
import subprocess

# Third-party library imports
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.manifold import MDS
from sklearn.metrics.pairwise import euclidean_distances

# Typing imports
from typing import Optional, Union

# Local application/library imports
from TINTOlib.abstractImageMethod import AbstractImageMethod
from TINTOlib.utils import Toolbox

###########################################################
################    REFINED    ##############################
###########################################################

class REFINED(AbstractImageMethod):
    """
    REFINED: A method to transform high-dimensional feature vectors into 2D images 
    optimized for CNNs. It uses Bayesian Metric Multidimensional Scaling (BMDS) 
    and hill climbing to arrange features spatially, preserving neighborhood dependencies.
    
    Parameters:
    ----------
    problem : str, optional
        The type of problem, defining how the images are grouped. 
        Default is 'supervised'. Valid values: ['supervised', 'unsupervised', 'regression'].
    normalize : bool, optional
        If True, normalizes input data using MinMaxScaler. 
        Default is True. Valid values: [True, False].
    verbose : bool, optional
        Show execution details in the terminal. 
        Default is False. Valid values: [True, False].
    hcIterations : int, optional
        Number of iterations for the hill climbing algorithm. 
        Default is 5. Valid values: integer >= 1.
    n_processors : int, optional    
        The number of processors to use for the algorithm. 
        Default is 8. Valid values: integer >= 2.
    zoom : int, optional
        Multiplication factor determining the size of the saved image relative to the original size. 
        Default is 1. Valid values: integer > 0.
    random_seed : int, optional
        Seed for reproducibility. 
        Default is 1. Valid values: integer.
    """
    ###### default values ###############
    default_hc_iterations = 5  # Number of iterations for the hill climbing algorithm
    default_n_processors = 8  # Default number of processors

    default_zoom = 1  # Default zoom level for saving images
    default_random_seed = 1  # Default seed for reproducibility

    def __init__(
        self,
        problem: Optional[str] = None,
        normalize: Optional[bool] = None,
        verbose: Optional[bool] = None,
        hcIterations: Optional[int] = default_hc_iterations,
        n_processors: Optional[int] = default_n_processors,
        zoom: Optional[int] = default_zoom,
        random_seed: Optional[int] = default_random_seed,
    ):   
        super().__init__(problem=problem, verbose=verbose, normalize=normalize)
        if n_processors < 2:
            raise ValueError(f"n_processors must be greater than 1 (got {n_processors})")
        
        self.hcIterations = hcIterations
        self.n_processors = n_processors

        self.zoom = zoom
        self.random_seed = random_seed

    def __saveSupervised(self, classValue, i, folder, matrix_a, fig, ax):
        extension = 'png'  # eps o pdf
        subfolder = str(int(classValue)).zfill(2)  # subfolder for grouping the results of each class
        name_image = str(i).zfill(6)
        route = os.path.join(folder, subfolder)
        route_complete = os.path.join(route, name_image + '.' + extension)
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        shape = int(math.sqrt(matrix_a.shape[0]))
        data = matrix_a.reshape(shape, shape)

        fig.set_size_inches(shape, shape)
        fig.set_dpi(self.zoom)
        ax.clear()
        plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
        ax.imshow(data, cmap='viridis', interpolation="nearest")
        ax.axis('off')
        fig.canvas.draw()
        fig.savefig(fname=route_complete, pad_inches=0, bbox_inches='tight', dpi=self.zoom)
        route_relative = os.path.join(subfolder, name_image+ '.' + extension)
        return route_relative

    def __saveRegressionOrUnsupervised(self, i, folder, matrix_a, fig, ax):
        extension = 'png'  # eps o pdf
        subfolder = "images"
        name_image = str(i).zfill(6)  + '.' + extension
        route = os.path.join(folder, subfolder)
        route_complete = os.path.join(route, name_image)

        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        shape = int(math.sqrt(matrix_a.shape[0]))
        data = matrix_a.reshape(shape,shape)

        fig.set_size_inches(shape, shape)
        fig.set_dpi(self.zoom)
        ax.clear()
        plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
        ax.imshow(data, cmap='viridis', interpolation="nearest")
        ax.axis('off')
        fig.canvas.draw()
        fig.savefig(fname=route_complete, pad_inches=0, bbox_inches='tight', dpi=self.zoom)
        route_relative = os.path.join(subfolder, name_image)
        return route_relative

    def __saveImages(self,gene_names,coords,map_in_int, X, Y, nn):

        gene_names_MDS, coords_MDS, map_in_int_MDS=(gene_names,coords,map_in_int)
        X_REFINED_MDS = Toolbox.REFINED_Im_Gen(X, nn, map_in_int_MDS, gene_names_MDS, coords_MDS)
        imagesRoutesArr=[]
        total = Y.shape[0]

        if self.verbose:
            print("SAVING")
        
        fig,ax = plt.subplots()

        for i in range(len(X_REFINED_MDS)):
            if self.problem == "supervised":
                route=self.__saveSupervised(Y[i], i, self.folder, X_REFINED_MDS[i], fig, ax)
                imagesRoutesArr.append(route)

            elif self.problem == "unsupervised" or self.problem == "regression" :
                route = self.__saveRegressionOrUnsupervised(i, self.folder, X_REFINED_MDS[i], fig, ax)
                imagesRoutesArr.append(route)
            else:
                print("Wrong problem definition. Please use 'supervised', 'unsupervised' or 'regression'")
            if self.verbose:
                print("Created ", str(i+1), "/", int(total))

        if self.problem == "supervised" :
            data={'images':imagesRoutesArr,'class':Y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/supervised.csv", index=False)
        elif self.problem == "unsupervised":
            data = {'images': imagesRoutesArr}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/unsupervised.csv", index=False)
        elif self.problem == "regression":
            data = {'images': imagesRoutesArr,'values':Y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/regression.csv", index=False)

    def _fitAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        Desc = x.columns.tolist()
        X = x.values
        Y = y.values if y is not None else None

        original_input = pd.DataFrame(data=X)  # The MDS input should be in a dataframe format with rows as samples and columns as features
        feature_names_list = original_input.columns.tolist()  # Extracting feature_names_list (gene_names or descriptor_names)
        if self.verbose:
            print(">>>> Data  is loaded")

        nn = math.ceil(np.sqrt(len(feature_names_list)))  # Image dimension
        Nn = original_input.shape[1]  # Number of features

        transposed_input = original_input.T  # The MDS input data must be transposed , because we want summarize each feature by two values (as compard to regular dimensionality reduction each sample will be described by two values)
        Euc_Dist = euclidean_distances(transposed_input)  # Euclidean distance
        Euc_Dist = np.maximum(Euc_Dist, Euc_Dist.transpose())  # Making the Euclidean distance matrix symmetric

        embedding = MDS(n_components=2, random_state=self.random_seed)  # Reduce the dimensionality by MDS into 2 components
        mds_xy = embedding.fit_transform(transposed_input)  # Apply MDS

        if self.verbose:
            print(">>>> MDS dimensionality reduction is done")

        eq_xy = Toolbox.two_d_eq(mds_xy, Nn)
        Img = Toolbox.Assign_features_to_pixels(eq_xy, nn,verbose=self.verbose)  # Img is the none-overlapping coordinates generated by MDS

        Desc = original_input.columns.tolist()                              # Drug descriptors name
        Dist = pd.DataFrame(data = Euc_Dist, columns = Desc, index = Desc)	# Generating a distance matrix which includes the Euclidean distance between each and every descriptor
        data = (Desc, Dist, Img	)  											# Preparing the hill climbing inputs

        init_pickle_file = "Init_MDS_Euc.pickle"
        with open(init_pickle_file, 'wb') as f:					# The hill climbing input is a pickle, therefore everything is saved as a pickle to be loaded by the hill climbing
            pickle.dump(data, f)

        mapping_pickle_file = "Mapping_REFINED_subprocess.pickle"
        evolution_csv_file = "REFINED_Evolve_subprocess.csv"
        script_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "utils","mpiHill_UF.py"
        )
        
        if 'Windows' == platform.system():
            command = f'mpiexec -np {self.n_processors} python {script_path} --init "{init_pickle_file}" --mapping "{mapping_pickle_file}"  --evolution "{evolution_csv_file}" --num {self.hcIterations}'
            result = subprocess.run(command, shell=True, text=True, capture_output=True)
        else:
            command = f'mpirun --allow-run-as-root --use-hwthread-cpus -np {self.n_processors} python3 {script_path} --init "{init_pickle_file}" --mapping "{mapping_pickle_file}"  --evolution "{evolution_csv_file}" --num {self.hcIterations}'
            result = subprocess.run(command, shell=True, text=True, capture_output=True)

        if result.returncode != 0:
            raise Exception(result.stderr)

        with open(mapping_pickle_file,'rb') as file:
            self.gene_names_MDS, self.coords_MDS, self.map_in_int_MDS = pickle.load(file)
    
        os.remove(init_pickle_file)
        os.remove(mapping_pickle_file)
        os.remove(evolution_csv_file)

    def _transformAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        X = x.values
        Y = y.values if y is not None else None

        original_input = pd.DataFrame(data=X)  # The MDS input should be in a dataframe format with rows as samples and columns as features
        feature_names_list = original_input.columns.tolist()  # Extracting feature_names_list (gene_names or descriptor_names)
        nn = math.ceil(np.sqrt(len(feature_names_list)))  # Image dimension
        self.__saveImages(self.gene_names_MDS, self.coords_MDS, self.map_in_int_MDS, X, Y, nn)


# Standard library imports
import os
import shutil

# Third-party library imports
import bitstring
import matplotlib
import matplotlib.image
import numpy as np
import pandas as pd

# Typing imports
from typing import Iterator, List, Union

# Local application/library imports
from TINTOlib.abstractImageMethod import AbstractImageMethod

###########################################################
################    Binary Image Encoding    ##############
###########################################################

default_precision = 32
default_zoom = 1

class BIE(AbstractImageMethod):
    """
    BIE: Generates 1-channel images by encoding the floating-point representation of numeric values.

    Each feature's value is converted to its binary representation, which is then used to create rows in the image. 
    This method preserves precise numeric information but may not effectively capture relationships between features.

    Parameters:
    ----------
    problem : str, optional
        The type of problem, defining how the images are grouped. 
        Default is 'supervised'. Valid values: ['supervised', 'unsupervised', 'regression'].
    normalize : bool, optional
        If True, normalizes input data using MinMaxScaler. 
        Default is True. Valid values: [True, False].
    verbose : bool, optional
        Show execution details in the terminal. 
        Default is False. Valid values: [True, False].
    precision : int, optional
        Determines the precision of the binary encoding. 
        Default is 32. Valid values: [32, 64].
    zoom : int, optional
        Multiplication factor determining the size of the saved image relative to the original size. 
        Default is 1. Valid values: integer > 0.
    """
    ###### Default values ###############
    default_precision = 32  # Precision for binary encoding
    default_zoom = 1  # Rescale factor for saving the image

    def __init__(
        self,
        problem = None,
        normalize=None,
        verbose = None,
        precision: int = default_precision,
        zoom: int = default_zoom
    ):
        super().__init__(problem=problem, verbose=verbose, normalize=normalize)

        if not isinstance(precision, int):
            raise TypeError(f"precision must be of type int (got {type(precision)})")
        configurable_precisions = [32, 64]
        if precision not in configurable_precisions:
            raise ValueError(f"precision must have one of this values {configurable_precisions}. Instead, got {precision}")
        
        if not isinstance(zoom, int):
            raise TypeError(f"zoom must be of type int (got {type(zoom)})")
        if zoom <= 0:
            raise ValueError(f"zoom must be positive. Instead, got {zoom}")
        
        self.precision = precision

        self.zoom = zoom
        
        self.ones, self.zeros = 255, 0

    def __convert_samples_to_binary(self, data: np.ndarray) -> Iterator[List[List[int]]]:
        def process_sample(sample):
            return [[self.ones if b=='1' else self.zeros for b in bitstring.BitArray(float=feat, length=self.precision).bin] for feat in sample]
        return map(process_sample, data)
    
    def __saveSupervised(self, matrix: np.ndarray, i: int, y):
        extension = 'png'
        subfolder = str(int(y)).zfill(2)  # subfolder for grouping the results of each class
        name_image = str(i).zfill(6)
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image + '.' + extension)

        if not os.path.exists(route):
            os.makedirs(route)

        matplotlib.image.imsave(route_complete, matrix, cmap='gray', format=extension, dpi=self.zoom, vmin=0, vmax=1)

        route_relative = os.path.join(subfolder, name_image+ '.' + extension)
        return route_relative

    def __saveRegressionOrUnsupervised(self, matrix: np.ndarray, i: int):
        extension = 'png'  # eps o pdf
        subfolder = "images"
        name_image = str(i).zfill(6) + '.' + extension
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image)

        if not os.path.exists(route):
            os.makedirs(route)

        matplotlib.image.imsave(route_complete, matrix, cmap='gray', format=extension, dpi=self.zoom, vmin=0, vmax=1)

        route_relative = os.path.join(subfolder, name_image)
        return route_relative

    def __save_images(self, matrices: Iterator[List[List[int]]], y, num_elems):
        imagesRoutesArr=[]

        if os.path.exists(self.folder):
            shutil.rmtree(self.folder)
        os.makedirs(self.folder)

        for (i,matrix) in enumerate(matrices):
            # Scale the matrix
            matrix = np.repeat(np.repeat(matrix, self.zoom, axis=0), self.zoom, axis=1)

            if self.problem == "supervised":
                route=self.__saveSupervised(matrix, i, y[i])
            elif self.problem == "unsupervised" or self.problem == "regression":
                route = self.__saveRegressionOrUnsupervised(matrix, i)
        
            imagesRoutesArr.append(route)

            if self.verbose:
                print("Created ", str(i + 1), "/", int(num_elems))
        
        if self.problem == "supervised":
            data = {'images': imagesRoutesArr, 'class': y}
            supervisedCSV = pd.DataFrame(data=data)
            supervisedCSV.to_csv(self.folder + "/supervised.csv", index=False)
        elif self.problem == "unsupervised":
            data = {'images': imagesRoutesArr}
            unsupervisedCSV = pd.DataFrame(data=data)
            unsupervisedCSV.to_csv(self.folder + "/unsupervised.csv", index=False)
        elif self.problem == "regression":
            data = {'images': imagesRoutesArr, 'values': y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/regression.csv", index=False)

    def _fitAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        """
        Fit method for stateless transformers. Does nothing and returns self.
        """
        return self
    
    def _transformAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        x = x.values
        y = y.values if y is not None else None

        matrices = self.__convert_samples_to_binary(x)
        self.__save_images(matrices, y, num_elems=x.shape[0])



# Standard library imports
import os

# Third-party library imports
import numpy as np
import pandas as pd
from PIL import Image

# Typing imports
from typing import Optional, Union

# Local application/library imports
from TINTOlib.abstractImageMethod import AbstractImageMethod

###########################################################
################    BarGraph    ###########################
###########################################################

class BarGraph(AbstractImageMethod):
    """
    BarGraph: Represents normalized variable values within [0, 1] in a bar graph format.

    This method visualizes data as a bar graph, where each variable is represented as a bar with a configurable 
    width and gap between bars.

    Parameters:
    ----------
    problem : str, optional
        The type of problem, defining how the images are grouped. 
        Default is 'supervised'. Valid values: ['supervised', 'unsupervised', 'regression'].
    normalize : bool, optional
        If True, normalizes input data using MinMaxScaler. 
        Default is True. Valid values: [True, False].
    verbose : bool, optional
        Show execution details in the terminal. 
        Default is False. Valid values: [True, False].
    pixel_width : int, optional
        The width of each bar in pixels. 
        Default is 1. Valid values: integer > 0.
    gap : int, optional
        The gap between bars in pixels. 
        Default is 0. Valid values: integer >= 0.
    zoom : int, optional
        Multiplication factor determining the size of the saved image relative to the original size. 
        Default is 1. Valid values: integer > 0.
    """
    ###### Default values ###############
    default_pixel_width = 1  # Width of the bars in pixels
    default_gap = 0  # Gap between the bars
    default_zoom = 1  # Rescale factor for saving the image

    def __init__(
        self,
        problem = None,
        normalize=None,
        verbose = None,
        pixel_width: int = default_pixel_width,
        gap: int = default_gap,
        zoom: int = default_zoom,
    ):
        super().__init__(problem=problem, verbose=verbose, normalize=normalize)

        if not isinstance(pixel_width, int):
            raise TypeError(f"pixel_width must be of type int (got {type(pixel_width)})")
        if pixel_width <= 0:
            raise ValueError(f"pixel_width must be positive (got {pixel_width})")
        if not isinstance(gap, int):
            raise TypeError(f"gap must be of type int (got {type(gap)})")
        if pixel_width < 0:
            raise ValueError(f"gap cannot be negative (got {gap})")
        
        self.pixel_width = pixel_width
        self.gap = gap

        self.zoom = zoom
                
    def __saveSupervised(self, y, i, image):
        extension = 'png'  # eps o pdf
        subfolder = str(int(y)).zfill(2)  # subfolder for grouping the results of each class
        name_image = str(i).zfill(6)
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image + '.' + extension)
        # Subfolder check
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        img = Image.fromarray(np.uint8(np.squeeze(image) * 255))
        img = img.resize(size=(img.size[0]*self.zoom, img.size[1]*self.zoom), resample=Image.Resampling.NEAREST)
        img.save(route_complete)

        route_relative = os.path.join(subfolder, name_image+ '.' + extension)
        return route_relative

    def __saveRegressionOrUnsupervised(self, i, image):
        extension = 'png'  # eps o pdf
        subfolder = "images"
        name_image = str(i).zfill(6) + '.' + extension
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image)
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        img = Image.fromarray(np.uint8(np.squeeze(image) * 255))
        img = img.resize(size=(img.size[0]*self.zoom, img.size[1]*self.zoom), resample=Image.Resampling.NEAREST)
        img.save(route_complete)

        route_relative = os.path.join(subfolder, name_image)
        return route_relative
    
    def _fitAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        """
        Fit method for stateless transformers. Does nothing and returns self.
        """
        return self
    
    def _transformAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        x = x.values
        Y = y.values if y is not None else None

        # TODO: reorder columns

        imagesRoutesArr = []    # List to store the routes

        # Image variables
        n_columns = x.shape[1]
        
        # There is a gap before the first column and after all the columns (n_columns columns & n_colums+1 gaps)
        image_size = self.pixel_width*n_columns + (n_columns+1)*self.gap
        # Add a padding on top and bottom
        top_padding, bottom_padding = self.pixel_width, self.pixel_width
        max_bar_height = image_size - (bottom_padding + top_padding)
        # Step of column (width + gap)
        step_column = self.gap + self.pixel_width

        for i,sample in enumerate(x):
            # Create the image (the image is squared)
            image = np.zeros([image_size, image_size, 1])
            # Multiply the values in the sample time the height of the bar
            bar_heights = np.floor(sample * max_bar_height).astype(np.int64)
            for i_bar,val_bar in enumerate(bar_heights):
                image[
                    top_padding : (top_padding + val_bar),                                         # The height of the column
                    (self.gap+(step_column*i_bar)) : (self.gap+(step_column*i_bar)) + self.pixel_width # The width of the column
                ] = 1
                
            if self.problem == "supervised":
                route = self.__saveSupervised(Y[i], i, image)
                imagesRoutesArr.append(route)
            elif self.problem == "unsupervised" or self.problem == "regression":
                route = self.__saveRegressionOrUnsupervised(i, image)
                imagesRoutesArr.append(route)
            else:
                print("Wrong problem definition. Please use 'supervised', 'unsupervised' or 'regression'")

        if self.problem == "supervised":
            data = {'images': imagesRoutesArr, 'class': Y}
            supervisedCSV = pd.DataFrame(data=data)
            supervisedCSV.to_csv(self.folder + "/supervised.csv", index=False)
        elif self.problem == "unsupervised":
            data = {'images': imagesRoutesArr}
            unsupervisedCSV = pd.DataFrame(data=data)
            unsupervisedCSV.to_csv(self.folder + "/unsupervised.csv", index=False)
        elif self.problem == "regression":
            data = {'images': imagesRoutesArr, 'values': Y}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/regression.csv", index=False)



# -*- coding: utf-8 -*-
"""
Created on Mon Oct  7 18:41:14 2019

@author: obazgir
"""
#FOR REFINED

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression




#%% NRMSE
def NRMSE(Y_Target, Y_Predict):
    Y_Target = np.array(Y_Target); Y_Predict = np.array(Y_Predict);
    Y_Target = Y_Target.reshape(len(Y_Target),1);    Y_Predict = Y_Predict.reshape(len(Y_Predict),1);    
    Y_Bar = np.mean(Y_Target)
    Nom = np.sum((Y_Predict - Y_Target)**2);    Denom = np.sum((Y_Bar - Y_Target)**2)
    MSE = np.mean((Y_Predict - Y_Target)**2);   NRMSE = np.sqrt(Nom/Denom)
    R2 = 1 - NRMSE**2
    return NRMSE, R2

def NMAE(Y_Target, Y_Predict):
    Y_Target = np.array(Y_Target); Y_Predict = np.array(Y_Predict);
    Y_Target = Y_Target.reshape(len(Y_Target),1);    Y_Predict = Y_Predict.reshape(len(Y_Predict),1);    
    Y_Bar = np.mean(Y_Target)
    Nom = np.abs((Y_Predict - Y_Target));    Denom = np.abs((Y_Bar - Y_Target))
    NormMAE = np.mean(Nom)/np.mean(Denom)
    return NormMAE
    

#%% Random position generation
import math
def Random_position(p):
    NN = int(math.sqrt(p)) +1
    Feat_num = np.arange(p)
    np.random.shuffle(Feat_num)
    #Feature_List = [f'F{i}' for i in Feat_num]
    Pos_mat = []
    for i in range(p):
        Pos_mat.append([int(Feat_num[i]/NN),int(Feat_num[i]%NN)])
    return(Pos_mat)
    
def Random_Image_Gen(X, Rand_Pos_mat):
    sz = X.shape
    p = sz[1];  N = sz[0]
    NN = int(math.sqrt(p)) +1
    Im = np.zeros((NN,NN))
    X_Gen = np.zeros((N,NN**2))
    for j in range(N):
        for i in range(p):
            P = Rand_Pos_mat[i]
            Im[P[0],P[1]] = X[j,i]
        Image_Store = Im.reshape((NN**2,1)).T
        X_Gen[j,:] = Image_Store
    return X_Gen

#%% MDS by Ruibo
from sklearn.manifold import MDS

def two_d_norm(xy):
    # xy is N x 2 xy cordinates, returns normed-xy on [0,1]
    norm_xy = (xy - xy.min(axis = 0)) / (xy - xy.min(axis = 0)).max(axis = 0)
    return norm_xy

def two_d_eq(xy,Nn):
    # xy is N x 2 xy cordinates, returns eq-xy on [0,1]
    xx_rank = np.argsort(xy[:,0])
    yy_rank = np.argsort(xy[:,1])
    eq_xy = np.full(xy.shape,np.nan)
    for ii in range(xy.shape[0]):
        xx_idx = xx_rank[ii]
        yy_idx = yy_rank[ii]
        eq_xy[xx_idx,0] = ii * 1/Nn
        eq_xy[yy_idx,1] = ii * 1/Nn
    return eq_xy

#embedding = MDS(n_components=2)
#mds_xy = embedding.fit_transform(transposed_input)
# to pixels
def Assign_features_to_pixels(xy,nn,verbose = False):
    # For each unassigned feature, find its nearest pixel, repeat until every ft is assigned
    # xy is the 2-d coordinates (normalized to [0,1]); nn is the image width. Img size = n x n
    # generate the result summary table, xy pixels; 3rd is nan for filling the idx
    Nn = xy.shape[0]
    
    from itertools import product
    pixel_iter = product([x for x in range(nn)],repeat = 2)
    result_table = np.full((nn*nn,3),np.nan)
    ii = 0
    for each_pixel in pixel_iter:
        result_table[ii,:2] = np.array(each_pixel)
        ii+=1
    # Use numpy array for speed 
        
    from sklearn.metrics import pairwise_distances
    
#    xy = eq_xy
    centroids = result_table[:,:2] / nn + 0.5/nn
    pixel_avail = np.ones(nn*nn).astype(bool)
    feature_assigned = np.zeros(Nn).astype(bool)
    
    dist_xy_centroids = pairwise_distances(centroids,xy,metric='euclidean')
    
    while feature_assigned.sum()<Nn:
        # Init the pick-relationship table
        pick_xy_centroids = np.zeros(dist_xy_centroids.shape).astype(bool)
        
        for each_feature in range(Nn):
            # for each feature, find the nearest available pixel
            if feature_assigned[each_feature] == True:
                # if this feature is already assigned, skip to the next ft
                continue
            else:
                # if not assigned:
                for ii in range(nn*nn):
                    # find the nearest avail pixel
                    nearest_pixel_idx = np.argsort(dist_xy_centroids[:,each_feature])[ii]
                    if pixel_avail[nearest_pixel_idx] == True:
                        break
                    else:
                        continue
                pick_xy_centroids[nearest_pixel_idx,each_feature] = True
            
        for each_pixel in range(nn*nn):
            # Assign the feature No to pixels
            if pixel_avail[each_pixel] == False:
                continue
            else:
                # find all the "True" features. np.where returns a tuple size 1
                related_features = np.where(pick_xy_centroids[each_pixel,:] == 1)[0]
                if len(related_features) == 1:
                    # Assign it
                    result_table[each_pixel,2] = related_features[0]
                    pixel_avail[each_pixel] = False
                    feature_assigned[related_features[0]] = True
                elif len(related_features) > 1:
                    related_dists = dist_xy_centroids[each_pixel,related_features]
                    best_feature = related_features[np.argsort(related_dists)[0]] # Sort, and pick the nearest one among them
                    result_table[each_pixel,2] = best_feature
                    pixel_avail[each_pixel] = False
                    feature_assigned[best_feature] = True
        if verbose:
            print(">> Assign features to pixels:", feature_assigned.sum(),"/",Nn)
    result_table[np.isnan(result_table[:, 2]), 2] = -1

    img = np.full((nn,nn),'NaN').astype(object)
    for each_pixel in range(nn*nn):
        xx = int(result_table[each_pixel,0])
        yy = int(result_table[each_pixel,1])
        ft = 'F' + str(int(result_table[each_pixel,2]))
        img[xx,yy] = ft
    return img.astype(object)

#eq_xy = two_d_eq(mds_xy)
#Img = Assign_features_to_pixels(eq_xy,nn,verbose=1)
#Init_Corr_MDS = InitCorr(dist_mat,Img,nn)

def MDS_Im_Gen(X,nn, Img):
    [N_sam,P_Feat] = X.shape
    X_Gen = np.zeros((N_sam,nn**2))
    conv_Img = Img.reshape(Img.size,1)
    for i in range(nn**2):
        Feature = np.array(conv_Img[i]);    Feature = Feature[0]; F_Num = int(Feature[1:])
        if abs(F_Num) < nn**2:
            X_Gen[:,i] = X[:,F_Num]
        else:
            X_Gen[:,i] = 0 
    return X_Gen

#%% CCLE functions
def dataframer(Main,Set_in, name_in, name_out):
    A = Set_in[name_in].tolist()
    Set_out = Main[Main[name_out] == A[0]]
    for cell in range(len(A) - 1):
        df = Main[Main[name_out] == A[cell + 1]]
        Set_out = pd.concat([Set_out, df])
    return Set_out

def Reg_to_Class(Y,Threshold):
	Y_Class = np.zeros(len(Y))
	Y_Sens = np.where(Y > Threshold)
	Y_Class[Y_Sens] = 1
	Y_Class =  Y_Class.astype(int)
	Y_Class = Y_Class.tolist()
	Y_Class = np.array(Y_Class)
	return Y_Class
	
def floattoint(Y_Test_Encoded):
	Y_Class = np.zeros(Y_Test_Encoded.shape)
	Y_Sens = np.where(Y_Test_Encoded > 0.5)
	Y_Class[Y_Sens] = 1
	Y_Class =  Y_Class.astype(int)
	Y_Class = Y_Class.tolist()
	Y_Class = np.array(Y_Class)
	return Y_Class
def REFINED_Im_Gen(X,nn, map_in_int, gene_names,coords):
	[N_sam,P_Feat] = X.shape
	X_Gen = np.zeros((N_sam,nn**2))
	for i in range(N_sam):
		data = X[i,:] 
		X_REFINED = pd.DataFrame(data = data.reshape(1,len(data)), columns = gene_names)
		Image = np.zeros(map_in_int.shape)
		for j in range(len(coords)):
			val = np.array(X_REFINED[gene_names[j]])
			Image[coords[j,0],coords[j,1]] = val
		Image = Image.reshape(nn**2)
		X_Gen[i,:] = Image
	return X_Gen
#%% GDSC
def GDSC_dataframer(PD_Set, Set_Name,PD_Attribute,Attribute_Name):
    A = PD_Set[Set_Name].tolist()
    b = PD_Attribute[PD_Attribute[Attribute_Name] == A[0]].reset_index().drop(columns = ['index'])
    Data_arry = np.array(b.values[0,1:],dtype = float)
    Data_arry = Data_arry.reshape(1,len(Data_arry))
    for i in range(len(A) - 1 ):
        b = PD_Attribute[PD_Attribute[Attribute_Name] == A[i + 1]].reset_index().drop(columns = ['index'])
        Arr = np.array(b.values[0,1:],dtype = float)
        Arr = Arr.reshape(1,len(Arr))
        Data_arry = np.append(Data_arry,Arr, axis = 0)
    
    PD_Data_arry = pd.DataFrame(data = Data_arry, columns = PD_Attribute.columns.tolist()[1:], index = A)
    return Data_arry, PD_Data_arry
	
def GDSC_NPier(PD_Set, Set_Name,PD_Attribute,Attribute_Name):
    PD_Set = PD_Set.reset_index()
    PD_Set.shape[0]
    PD_Attribute.shape[1] - 1
    X_NP = np.zeros((PD_Set.shape[0],PD_Attribute.shape[1] - 1))
    Source = list(set(PD_Set[Set_Name].tolist()))
    for name in Source:
        idx = PD_Set.index[PD_Set[Set_Name] == name].tolist()
        XX = np.array(PD_Attribute[PD_Attribute[Attribute_Name] == name].values[0,1:], dtype = float)
        X_NP[idx,:] = XX
    return X_NP
	
	
def Coord_Converter(coords_drug2,nn):
    coords_drug3 = np.full((nn,nn),'NaN').astype(object)
    for i in range(nn):
        for j in range(nn):
            ft = 'F' + str(coords_drug2[i,j])
            coords_drug3[i,j] = ft
    return coords_drug3
	
def Bias_Calc(Y_Test, Y_Pred):
    Error = Y_Test - Y_Pred
    Y_Test = Y_Test.reshape(len(Y_Test),1)
    Error = Error.reshape(len(Error),1)
    
    reg = LinearRegression().fit(Y_Test, Error)
    Bias = reg.coef_[0]
    
    return Bias
	



# Standard library imports
import pickle
from abc import ABC, abstractmethod

# Third-party library imports
import pandas as pd
from sklearn.preprocessing import MinMaxScaler

# Typing imports
from typing import Optional, Union

# Default configuration values
default_problem = "supervised"  # Define the type of task [supervised, unsupervised, regression]
default_verbose = False         # Verbose: if True, shows the compilation text
default_normalize = True
default_hyperparameters_filename = 'objs.pkl'

class AbstractImageMethod(ABC):
    """
    Abstract class that other classes must inherit from and implement abstract methods.
    Provides utility methods for saving/loading hyperparameters and data transformations.
    """
    def __init__(
        self,
        problem: Optional[str], 
        verbose: Optional[bool],
        normalize: Optional[bool],
    ):
        # Validate `problem`
        if problem is None:
            problem = default_problem
        allowed_values_for_problem = ["supervised", "unsupervised", "regression"]
        if not isinstance(problem, str):
            raise TypeError(f"problem must be of type str (got {type(problem)})")
        if problem not in allowed_values_for_problem:
            raise ValueError(f"Allowed values for problem are {allowed_values_for_problem}. Instead got {problem}")
        
        # Validate `verbose`
        if verbose is None:
            verbose = default_verbose
        if not isinstance(verbose, bool):
            raise TypeError(f"verbose must be of type bool (got {type(verbose)})")
        
        # Validate `normalize`
        if normalize is None:
            normalize = default_normalize
        if not isinstance(normalize, bool):
            raise TypeError(f"verbose must be of type bool (got {type(normalize)})")

        self.problem = problem
        self.verbose = verbose
        self._fitted = False  # Tracks if fit has been called

        # Normalize data
        self.normalize = normalize  # Whether to normalize data
        self.scaler = MinMaxScaler() if normalize else None  # Initialize scaler if needed


    def saveHyperparameters(self, filename=default_hyperparameters_filename):
        """
        This function allows SAVING the transformation options to images in a Pickle object.
        This point is basically to be able to reproduce the experiments or reuse the transformation
        on unlabelled data.
        """
        with open(filename, 'wb') as f:
            pickle.dump(self.__dict__, f)
        if self.verbose:
            print(f"Hyperparameters successfully saved in {filename}.")

    def loadHyperparameters(self, filename=default_hyperparameters_filename):
        """
        This function allows LOADING the transformation options to images from a Pickle object.
        This point is basically to be able to reproduce the experiments or reuse the transformation
        on unlabelled data.
        """
        with open(filename, 'rb') as f:
            variables = pickle.load(f)
        
        for key, val in variables.items():
            setattr(self, key, val)

        if self.verbose:
            print(f"Hyperparameters successfully loaded from {filename}.")
        
    def fit(self, data):
        """
        Fits the model to the tabular data.

        Parameters:
        - data: Path to CSV file or a pandas DataFrame containing data and targets.
        """

        dataset = self._load_data(data)
        x, y = self._split_features_targets(dataset)

        # Normalize features if required
        if self.normalize:
            x = pd.DataFrame(self.scaler.fit_transform(x), columns=x.columns)

        # Call the training function
        self._fitAlg(x, y)

        self._fitted = True  # Mark as fitted

        if self.verbose:
            print("Fit process completed.")

    def transform(self, data, folder):
        """
        Generate and saves the synthetic images in the specified folder.
        
        Parameters:
        - data: Path to CSV file or a pandas DataFrame containing data and targets.
        - folder: Path to folder where the images will be saved.
        """
        if not self._fitted:
            raise RuntimeError("The model must be fitted before calling 'transform'. Please call 'fit' first.")

        dataset = self._load_data(data)
        x, y = self._split_features_targets(dataset)

        # Normalize features if required
        if self.normalize:
            x = pd.DataFrame(self.scaler.transform(x), columns=x.columns)

        self.folder = folder
        self._transformAlg(x, y)

        if self.verbose:
            print("Transform process completed.")

    def fit_transform(self, data, folder):
        """
        Fits the model to the tabular data and then generate and saves the synthetic images in the specified folder.
        
        Parameters:
        - data: Path to CSV file or a pandas DataFrame containing data and targets.
        - folder: Path to folder where the images will be saved.
        """
        dataset = self._load_data(data)
        x, y = self._split_features_targets(dataset)

        # Normalize features if required
        if self.normalize:
            x = pd.DataFrame(self.scaler.fit_transform(x), columns=x.columns)

        self.folder = folder
        self._fitAlg(x, y)
        self._transformAlg(x, y)
        self._fitted = True  # Mark as fitted after both operations
    
        if self.verbose:
            print("Fit-Transform process completed.")

    def _load_data(self, data: Union[str, pd.DataFrame]) -> pd.DataFrame:
        """
        Loads data from a file or returns the DataFrame directly.
        """
        if isinstance(data, str):
            dataset = pd.read_csv(data)
        elif isinstance(data, pd.DataFrame):
            dataset = data
        else:
            raise TypeError("data must be a string (file path) or a pandas DataFrame.")

        if self.verbose:
            print("Data successfully loaded.")

        return dataset

    def _split_features_targets(self, dataset: pd.DataFrame):
        """
        Splits dataset into features and targets based on the problem type.
        """
        if self.problem in ["supervised", "regression"]:
            x = dataset.drop(columns=dataset.columns[-1])
            y = dataset[dataset.columns[-1]]
        else:
            x = dataset
            y = None

        return x, y

    @abstractmethod
    def _fitAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        """
        Abstract method for fitting the algorithm. Must be implemented by subclasses.
        This method is not to be called from the outside.
        """
        raise NotImplementedError("Subclasses must implement _fit_alg.")

    @abstractmethod
    def _transformAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        """
        Abstract method for transforming the data. Must be implemented by subclasses.
        This method is not to be called from the outside.
        """
        raise NotImplementedError("Subclasses must implement _transform_alg.")
    


# Standard library imports
import os
import shutil
import time

# Third-party library imports
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.spatial.distance import pdist, squareform
from scipy.stats import rankdata, spearmanr

# Typing imports
from typing import List, Optional, Union

# Local application/library imports
from TINTOlib.abstractImageMethod import AbstractImageMethod

###########################################################
################    IGTD    ###############################
###########################################################

class IGTD(AbstractImageMethod):
    """
    IGTD: Transforms tabular data into compact 2D images for CNNs by assigning features to pixels 
    based on similarity. It minimizes feature-pixel ranking differences, preserving neighborhood 
    structures effectively.

    Parameters:
    ----------
    problem : str, optional
        The type of problem, defining how the images are grouped. 
        Default is 'supervised'. Valid values: ['supervised', 'unsupervised', 'regression'].
    normalize : bool, optional
        If True, normalizes input data using MinMaxScaler. 
        Default is True. Valid values: [True, False].
    verbose : bool, optional
        Show execution details in the terminal. 
        Default is False. Valid values: [True, False].
    scale : List[int], optional
        The number of pixel rows and columns in the image. The product (rows x columns) must 
        be greater than or equal to the number of features. 
        Default is [6,6]. Valid values: list of two positive integers.
    fea_dist_method : str, optional
        Method to calculate pairwise distances between features.
        Default is 'Pearson'. Valid values: ['Pearson', 'Spearman', 'set', 'Euclidean'].
    image_dist_method : str, optional
        Method to calculate distances between pixels in the image.
        Default is 'Euclidean'. Valid values: ['Euclidean', 'Manhattan'].
    error : str, optional
        Function to evaluate differences between feature and pixel distance rankings.
        Default is 'squared'. Valid values: ['squared', 'abs'].
    max_step : int, optional
        Maximum number of iterations for the algorithm if it does not converge. 
        Default is 1000. Valid values: integer.
    val_step : int, optional
        Number of steps to check gain on the objective function for convergence. 
        Default is 50. Valid values: integer.
    switch_t : int, optional
        Threshold for error change rate to determine if switching features should occur. 
        Default is 0. Valid values: integer.
    min_gain : float, optional
        Minimum improvement in the objective function to continue optimization. 
        Default is 0.00001. Valid values: float.
    zoom : int, optional
        Multiplication factor determining the size of the saved image relative to the original size. 
        Default is 1. Valid values: integer > 0.
    random_seed : int, optional
        Seed for reproducibility. 
        Default is 1. Valid values: integer.    
    """
    #Default hyperparameters
    default_scale = [6,6]                       # Characteristic pixels of the final image (row x col). [row x col] must be equal or greater than the number of features.
    
    default_fea_dist_method = "Pearson"         # 'Pearson' uses Pearson correlation coefficient to evaluate similarity between features;
                                                # 'Spearman' uses Spearman correlation coefficient to evaluate similarity between features;
                                                # 'set' uses Jaccard index to evaluate similarity between features that are binary variables;
                                                # 'Euclidean' calculates pairwise euclidean distances between features.
    default_image_dist_method = "Euclidean"     # method used to calculate distance. Can be 'Euclidean' or 'Manhattan'.
    default_error = "squared"                   # a string indicating the function to evaluate the difference between feature distance ranking and pixel distance ranking. 'abs' indicates the absolute function. 'squared' indicates the square function

    default_max_step = 1000                     # the maximum steps that the algorithm should run if never converges.
    default_val_step = 50                       # number of steps for checking gain on the objective function to determine convergence
    default_switch_t = 0                        # the threshold to determine whether switch should happen
    default_min_gain = 0.00001                  # if the objective function is not improved more than 'min_gain' in 'val_step' steps, the algorithm terminates.
    
    default_zoom = 1                            # The default multiplication value to save the image
    default_random_seed = 1                     # default seed for reproducibility

    def __init__(
        self,
        problem: Optional[str] = None,
        normalize: Optional[bool] = None,
        verbose: Optional[bool] = None,
        scale: Optional[List[int]] = default_scale,
        fea_dist_method: Optional[str] = default_fea_dist_method,
        image_dist_method: Optional[str] = default_image_dist_method,
        error: Optional[str] = default_error,
        max_step: Optional[int] = default_max_step,
        val_step: Optional[int] = default_val_step,
        switch_t: Optional[int] = default_switch_t,
        min_gain: Optional[float] = default_min_gain,
        zoom: Optional[int] = default_zoom,
        random_seed: Optional[int] = default_random_seed,
    ):
        super().__init__(problem=problem, verbose=verbose, normalize=normalize)

        self.scale = scale
        self.fea_dist_method = fea_dist_method
        self.image_dist_method = image_dist_method
        self.zoom = zoom
        self.max_step = max_step
        self.val_step = val_step
        self.error = error
        self.switch_t = switch_t
        self.min_gain = min_gain
        self.random_seed = random_seed

    def __min_max_transform(self, data: np.ndarray):
        '''
        This function does a linear transformation of each feature, so that the minimum and maximum values of a
        feature are 0 and 1, respectively.

        Input
        -----
        data: ndarray
            an input data array with a size of [n_sample, n_feature]

        Return
        ------
        norm_data: ndarray
            the data array after transformation
        '''

        norm_data = np.empty(data.shape)
        norm_data.fill(np.nan)
        for i in range(data.shape[1]):
            v = data[:, i].copy()
            if np.max(v) == np.min(v):
                norm_data[:, i] = 0
            else:
                v = (v - np.min(v)) / (np.max(v) - np.min(v))
                norm_data[:, i] = v
        return norm_data

    def __generate_feature_distance_ranking(self, data:np.ndarray):
        '''
        This function generates ranking of distances/dissimilarities between features for tabular data.

        Input
        -----
        data: np.ndarray
            input data, n_sample by n_feature

        Return
        ------
        ranking: nd.ndarray
            symmetric ranking matrix based on dissimilarity
        corr:
            matrix of distances between features
        '''

        num = data.shape[1]
        if self.fea_dist_method == 'Pearson':
            corr = np.corrcoef(np.transpose(data))
        elif self.fea_dist_method == 'Spearman':
            corr = spearmanr(data).correlation
        elif self.fea_dist_method == 'Euclidean':
            corr = squareform(pdist(np.transpose(data), metric='euclidean'))
            corr = np.max(corr) - corr
            corr = corr / np.max(corr)
        elif self.fea_dist_method == 'set':  # This is the new set operation to calculate similarity. It does not tolerate all-zero features.
            corr1 = np.dot(np.transpose(data), data)
            corr2 = data.shape[0] - np.dot(np.transpose(1 - data), 1 - data)
            corr = corr1 / corr2

        corr = 1 - corr
        corr = np.around(a=corr, decimals=10)

        tril_id = np.tril_indices(num, k=-1)
        rank = rankdata(corr[tril_id])
        ranking = np.zeros((num, num))
        ranking[tril_id] = rank
        ranking = ranking + np.transpose(ranking)

        return ranking, corr

    def __generate_matrix_distance_ranking(self, num_r: int, num_c: int, num: int):
        '''
        This function calculates the ranking of distances between all pairs of entries in a matrix of size num_r by num_c.

        Input
        -----
        num_r: int
            number of rows in the matrix
        num_c: int
            number of columns in the matrix
        num: int

        Return
        ------
        coordinate: np.ndarray
            num_r * num_c by 2 matrix giving the coordinates of elements in the matrix.
        ranking: np.ndarray
            a num_r * num_c by num_r * num_c matrix giving the ranking of pair-wise distance.
        '''

        # generate the coordinates of elements in a matrix
        for r in range(num_r):
            if r == 0:
                coordinate = np.transpose(np.vstack((np.zeros(num_c), range(num_c))))
            else:
                coordinate = np.vstack((coordinate, np.transpose(np.vstack((np.ones(num_c) * r, range(num_c))))))
        coordinate = coordinate[:num, :]

        # calculate the closeness of the elements
        cord_dist = np.zeros((num, num))
        if self.image_dist_method == 'Euclidean':
            for i in range(num):
                cord_dist[i, :] = np.sqrt(np.square(coordinate[i, 0] * np.ones(num) - coordinate[:, 0]) +
                                          np.square(coordinate[i, 1] * np.ones(num) - coordinate[:, 1]))
        elif self.image_dist_method == 'Manhattan':
            for i in range(num):
                cord_dist[i, :] = np.abs(coordinate[i, 0] * np.ones(num) - coordinate[:, 0]) + \
                                  np.abs(coordinate[i, 1] * np.ones(num) - coordinate[:, 1])

        # generate the ranking based on distance
        tril_id = np.tril_indices(num, k=-1)
        rank = rankdata(cord_dist[tril_id])
        ranking = np.zeros((num, num))
        ranking[tril_id] = rank
        ranking = ranking + np.transpose(ranking)

        coordinate = np.int64(coordinate)
        return (coordinate[:, 0], coordinate[:, 1]), ranking

    def __IGTD_absolute_error(self, source, target):
        '''
        This function switches the order of rows (columns) in the source ranking matrix to make it similar to the target
        ranking matrix. In each step, the algorithm randomly picks a row that has not been switched with others for
        the longest time and checks all possible switch of this row, and selects the switch that reduces the
        dissimilarity most. Dissimilarity (i.e. the error) is the summation of absolute difference of
        lower triangular elements between the rearranged source ranking matrix and the target ranking matrix.

        Input
        -----
        source:
            a symmetric ranking matrix with zero diagonal elements.
        target:
            a symmetric ranking matrix with zero diagonal elements with the same size as 'source'.

        Return
        ------
        index_record:
            indices to rearrange the rows(columns) in source obtained the optimization process
        err_record:
            error obtained in the optimization process
        run_time:
            the time at which each step is completed in the optimization process
        '''

        r = np.random.RandomState(seed=self.random_seed)

        source = source.copy()
        num = source.shape[0]
        tril_id = np.tril_indices(num, k=-1)
        index = np.array(range(num))
        index_record = np.empty((self.max_step + 1, num))
        index_record.fill(np.nan)
        index_record[0, :] = index.copy()

        # calculate the error associated with each row
        err_v = np.empty(num)
        err_v.fill(np.nan)
        for i in range(num):
            err_v[i] = np.sum(np.abs(source[i, 0:i] - target[i, 0:i])) + \
                       np.sum(np.abs(source[(i + 1):, i] - target[(i + 1):, i]))

        step_record = -np.ones(num)
        err_record = [np.sum(abs(source[tril_id] - target[tril_id]))]
        pre_err = err_record[0]
        t1 = time.time()
        run_time = [0]

        for s in range(self.max_step):
            delta = np.ones(num) * np.inf

            # randomly pick a row that has not been considered for the longest time
            idr = np.where(step_record == np.min(step_record))[0]
            ii = idr[r.permutation(len(idr))[0]]

            for jj in range(num):
                if jj == ii:
                    continue

                if ii < jj:
                    i = ii
                    j = jj
                else:
                    i = jj
                    j = ii

                err_ori = err_v[i] + err_v[j] - np.abs(source[j, i] - target[j, i])

                err_i = np.sum(np.abs(source[j, :i] - target[i, :i])) + \
                        np.sum(np.abs(source[(i + 1):j, j] - target[(i + 1):j, i])) + \
                        np.sum(np.abs(source[(j + 1):, j] - target[(j + 1):, i])) + np.abs(source[i, j] - target[j, i])
                err_j = np.sum(np.abs(source[i, :i] - target[j, :i])) + \
                        np.sum(np.abs(source[i, (i + 1):j] - target[j, (i + 1):j])) + \
                        np.sum(np.abs(source[(j + 1):, i] - target[(j + 1):, j])) + np.abs(source[i, j] - target[j, i])
                err_test = err_i + err_j - np.abs(source[i, j] - target[j, i])

                delta[jj] = err_test - err_ori

            delta_norm = delta / pre_err
            id = np.where(delta_norm <= self.switch_t)[0]
            if len(id) > 0:
                jj = np.argmin(delta)

                # Update the error associated with each row
                if ii < jj:
                    i = ii
                    j = jj
                else:
                    i = jj
                    j = ii
                for k in range(num):
                    if k < i:
                        err_v[k] = err_v[k] - np.abs(source[i, k] - target[i, k]) - np.abs(source[j, k] - target[j, k]) + \
                                   np.abs(source[j, k] - target[i, k]) + np.abs(source[i, k] - target[j, k])
                    elif k == i:
                        err_v[k] = np.sum(np.abs(source[j, :i] - target[i, :i])) + \
                                   np.sum(np.abs(source[(i + 1):j, j] - target[(i + 1):j, i])) + \
                                   np.sum(np.abs(source[(j + 1):, j] - target[(j + 1):, i])) + np.abs(
                            source[i, j] - target[j, i])
                    elif k < j:
                        err_v[k] = err_v[k] - np.abs(source[k, i] - target[k, i]) - np.abs(source[j, k] - target[j, k]) + \
                                   np.abs(source[k, j] - target[k, i]) + np.abs(source[i, k] - target[j, k])
                    elif k == j:
                        err_v[k] = np.sum(np.abs(source[i, :i] - target[j, :i])) + \
                                   np.sum(np.abs(source[i, (i + 1):j] - target[j, (i + 1):j])) + \
                                   np.sum(np.abs(source[(j + 1):, i] - target[(j + 1):, j])) + np.abs(
                            source[i, j] - target[j, i])
                    else:
                        err_v[k] = err_v[k] - np.abs(source[k, i] - target[k, i]) - np.abs(source[k, j] - target[k, j]) + \
                                   np.abs(source[k, j] - target[k, i]) + np.abs(source[k, i] - target[k, j])

                # switch rows i and j
                ii_v = source[ii, :].copy()
                jj_v = source[jj, :].copy()
                source[ii, :] = jj_v
                source[jj, :] = ii_v
                ii_v = source[:, ii].copy()
                jj_v = source[:, jj].copy()
                source[:, ii] = jj_v
                source[:, jj] = ii_v
                err = delta[jj] + pre_err

                # update rearrange index
                t = index[ii]
                index[ii] = index[jj]
                index[jj] = t

                # update step record
                step_record[ii] = s
                step_record[jj] = s
            else:
                # error is not changed due to no switch
                err = pre_err

                # update step record
                step_record[ii] = s

            err_record.append(err)
            if self.verbose:
                print('Step ' + str(s) + ' err: ' + str(err))
            index_record[s + 1, :] = index.copy()
            run_time.append(time.time() - t1)

            if s > self.val_step:
                if np.sum((err_record[-self.val_step - 1] - np.array(err_record[(-self.val_step):])) / err_record[
                    -self.val_step - 1] >= self.min_gain) == 0:
                    break

            pre_err = err
        index_record = index_record[:len(err_record), :].astype(int)
        return index_record, err_record, run_time

    def __IGTD_square_error(self, source, target):
        '''
        This function switches the order of rows (columns) in the source ranking matrix to make it similar to the target
        ranking matrix. In each step, the algorithm randomly picks a row that has not been switched with others for
        the longest time and checks all possible switch of this row, and selects the switch that reduces the
        dissimilarity most. Dissimilarity (i.e. the error) is the summation of squared difference of
        lower triangular elements between the rearranged source ranking matrix and the target ranking matrix.

        Input
        -----
        source:
            a symmetric ranking matrix with zero diagonal elements.
        target:
            a symmetric ranking matrix with zero diagonal elements. 'source' and 'target' should have the same size.

        Return
        ------
        index_record:
            ordering index to rearrange the rows (columns) in 'source' in the optimization process
        err_record:
            the error history in the optimization process
        run_time:
            the time at which each step is finished in the optimization process
        '''

        r = np.random.RandomState(seed=self.random_seed)

        source = source.copy()
        num = source.shape[0]
        tril_id = np.tril_indices(num, k=-1)
        index = np.array(range(num))
        index_record = np.empty((self.max_step + 1, num))
        index_record.fill(np.nan)
        index_record[0, :] = index.copy()

        # calculate the error associated with each row
        err_v = np.empty(num)
        err_v.fill(np.nan)
        for i in range(num):
            err_v[i] = np.sum(np.square(source[i, 0:i] - target[i, 0:i])) + \
                       np.sum(np.square(source[(i + 1):, i] - target[(i + 1):, i]))

        step_record = -np.ones(num)
        err_record = [np.sum(np.square(source[tril_id] - target[tril_id]))]
        pre_err = err_record[0]
        t1 = time.time()
        run_time = [0]

        for s in range(self.max_step):
            delta = np.ones(num) * np.inf

            # randomly pick a row that has not been considered for the longest time
            idr = np.where(step_record == np.min(step_record))[0]
            ii = idr[r.permutation(len(idr))[0]]

            for jj in range(num):
                if jj == ii:
                    continue

                if ii < jj:
                    i = ii
                    j = jj
                else:
                    i = jj
                    j = ii

                err_ori = err_v[i] + err_v[j] - np.square(source[j, i] - target[j, i])

                err_i = np.sum(np.square(source[j, :i] - target[i, :i])) + \
                        np.sum(np.square(source[(i + 1):j, j] - target[(i + 1):j, i])) + \
                        np.sum(np.square(source[(j + 1):, j] - target[(j + 1):, i])) + np.square(
                    source[i, j] - target[j, i])
                err_j = np.sum(np.square(source[i, :i] - target[j, :i])) + \
                        np.sum(np.square(source[i, (i + 1):j] - target[j, (i + 1):j])) + \
                        np.sum(np.square(source[(j + 1):, i] - target[(j + 1):, j])) + np.square(
                    source[i, j] - target[j, i])
                err_test = err_i + err_j - np.square(source[i, j] - target[j, i])

                delta[jj] = err_test - err_ori

            delta_norm = delta / pre_err
            id = np.where(delta_norm <= self.switch_t)[0]
            if len(id) > 0:
                jj = np.argmin(delta)

                # Update the error associated with each row
                if ii < jj:
                    i = ii
                    j = jj
                else:
                    i = jj
                    j = ii
                for k in range(num):
                    if k < i:
                        err_v[k] = err_v[k] - np.square(source[i, k] - target[i, k]) - np.square(
                            source[j, k] - target[j, k]) + \
                                   np.square(source[j, k] - target[i, k]) + np.square(source[i, k] - target[j, k])
                    elif k == i:
                        err_v[k] = np.sum(np.square(source[j, :i] - target[i, :i])) + \
                                   np.sum(np.square(source[(i + 1):j, j] - target[(i + 1):j, i])) + \
                                   np.sum(np.square(source[(j + 1):, j] - target[(j + 1):, i])) + np.square(
                            source[i, j] - target[j, i])
                    elif k < j:
                        err_v[k] = err_v[k] - np.square(source[k, i] - target[k, i]) - np.square(
                            source[j, k] - target[j, k]) + \
                                   np.square(source[k, j] - target[k, i]) + np.square(source[i, k] - target[j, k])
                    elif k == j:
                        err_v[k] = np.sum(np.square(source[i, :i] - target[j, :i])) + \
                                   np.sum(np.square(source[i, (i + 1):j] - target[j, (i + 1):j])) + \
                                   np.sum(np.square(source[(j + 1):, i] - target[(j + 1):, j])) + np.square(
                            source[i, j] - target[j, i])
                    else:
                        err_v[k] = err_v[k] - np.square(source[k, i] - target[k, i]) - np.square(
                            source[k, j] - target[k, j]) + \
                                   np.square(source[k, j] - target[k, i]) + np.square(source[k, i] - target[k, j])

                # switch rows i and j
                ii_v = source[ii, :].copy()
                jj_v = source[jj, :].copy()
                source[ii, :] = jj_v
                source[jj, :] = ii_v
                ii_v = source[:, ii].copy()
                jj_v = source[:, jj].copy()
                source[:, ii] = jj_v
                source[:, jj] = ii_v
                err = delta[jj] + pre_err

                # update rearrange index
                t = index[ii]
                index[ii] = index[jj]
                index[jj] = t

                # update step record
                step_record[ii] = s
                step_record[jj] = s
            else:
                # error is not changed due to no switch
                err = pre_err

                # update step record
                step_record[ii] = s

            err_record.append(err)
            if self.verbose:
                print('Step ' + str(s) + ' err: ' + str(err))
            index_record[s + 1, :] = index.copy()
            run_time.append(time.time() - t1)

            if s > self.val_step:
                if np.sum((err_record[-self.val_step - 1] - np.array(err_record[(-self.val_step):])) / err_record[
                    -self.val_step - 1] >= self.min_gain) == 0:
                    break

            pre_err = err

        index_record = index_record[:len(err_record), :].astype(int)

        return index_record, err_record, run_time

    def __training(self, source, target):
        '''
        This is just a wrapper function that wraps the two search functions using different error measures.
        '''
        if self.error  == 'abs':
            index_record, err_record, run_time = self.__IGTD_absolute_error(source=source, target=target)
        if self.error == 'squared':
            index_record, err_record, run_time = self.__IGTD_square_error(source=source, target=target)

        return index_record, err_record, run_time

    def __saveSupervised(self, y, i, data_i, fig, ax):
        '''
        Saves the matrix as an image in a supervised dataset.

        Input
        -----
        i: int
            the index of the row within the dataset
        y:
            the true label of the i-th row
        data_i: ndarray
            the matrix containing the data to be saved as an image
        '''
        extension = 'png'  # eps o pdf
        subfolder = str(int(y)).zfill(2)  # subfolder for grouping the results of each class
        name_image = str(i).zfill(6) + '.' + extension
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image)
        # Subfolder check
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        fig.set_size_inches(self.scale[1], self.scale[0])
        fig.set_dpi(self.zoom)
        ax.clear()
        plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
        ax.imshow(data_i, cmap='gray', vmin=0, vmax=255, interpolation="nearest")
        ax.axis('off')
        fig.canvas.draw()
        fig.savefig(fname=route_complete, pad_inches=0, bbox_inches='tight', dpi=self.zoom)
        route_relative = os.path.join(subfolder, name_image)
        return route_relative

    def __saveRegressionOrUnsupervised(self, i, data_i, fig, ax):
        '''
        Saves the matrix as an image in a regression or unsupervised dataset.

        Input
        -----
        i: int
            the index of the row within the dataset
        data_i: ndarray
            the matrix containing the data to be saved as an image
        '''
        extension = 'png'  # eps o pdf
        subfolder = "images"
        name_image = str(i).zfill(6) + '.' + extension
        route = os.path.join(self.folder, subfolder)
        route_complete = os.path.join(route, name_image)
        if not os.path.isdir(route):
            try:
                os.makedirs(route)
            except:
                print("Error: Could not create subfolder")

        fig.set_size_inches(self.scale[1], self.scale[0])
        fig.set_dpi(self.zoom)
        ax.clear()
        plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
        ax.imshow(data_i, cmap='gray', vmin=0, vmax=255, interpolation="nearest")
        ax.axis('off')
        fig.canvas.draw()
        fig.savefig(fname=route_complete, pad_inches=0, bbox_inches='tight', dpi=self.zoom)
        route_relative = os.path.join(subfolder, name_image)
        return route_relative

    def __generate_image_data(self, data, index, num_row, num_column, coord, labels):
        '''
        This function generates the data in image format according to rearrangement indices. It saves the data
        sample-by-sample in both txt files and image files.

        Input
        -----
        data:
            original tabular data, 2D array or data frame, n_samples by n_features.
        index:
            indices of features obtained through optimization, according to which the features can be
            arranged into a num_r by num_c image.
        num_row: int
            number of rows in image.
        num_column: int
            number of columns in image.
        coord:
            coordinates of features in the image/matrix.
        labels: ndarray
            ndarray containing the labels for the tabular data

        Return
        ------
        image_data:
            the generated data, a 3D numpy array. The third dimension is across samples. The range of values
            is [0, 255]. Small values actually indicate high values in the original data.
        samples:
            the names of indices of the samples.
        '''
        image_folder = self.folder
        imagesRoutesArr = []
        if isinstance(data, pd.DataFrame):
            samples = data.index.map(np.str)
            data = data.values
        else:
            samples = [str(i) for i in range(data.shape[0])]

        if os.path.exists(image_folder):
            shutil.rmtree(image_folder)
        os.makedirs(image_folder)

        data_2 = data.copy()
        data_2 = data_2[:, index]
        max_v = np.max(data_2)
        min_v = np.min(data_2)
        data_2 = 255 - (data_2 - min_v) / (max_v - min_v) * 255  # So that black means high value

        image_data = np.empty((num_row, num_column, data_2.shape[0]))
        image_data.fill(np.nan)
        total = data_2.shape[0]

        fig,ax = plt.subplots()

        for i in range(data_2.shape[0]):
            data_i = np.empty((num_row, num_column))
            data_i.fill(np.nan)
            data_i[coord] = data_2[i, :]

            iid = np.where(np.isnan(data_i))
            data_i[iid] = 255

            image_data[:, :, i] = data_i
            image_data[:, :, i] = 255 - image_data[:, :, i]
            if image_folder is not None:
                if self.problem == "supervised":
                    route = self.__saveSupervised(labels[i], i, data_i, fig, ax)
                    imagesRoutesArr.append(route)
                elif self.problem == "unsupervised" or self.problem == "regression":
                    route = self.__saveRegressionOrUnsupervised(i, data_i, fig, ax)
                    imagesRoutesArr.append(route)
                else:
                    print("Wrong problem definition. Please use 'supervised', 'unsupervised' or 'regression'")

                # Verbose
                if self.verbose:
                    print("Created ", str(i + 1), "/", int(total))

        if self.problem == "supervised":
            data = {'images': imagesRoutesArr, 'class': labels}
            supervisedCSV = pd.DataFrame(data=data)
            supervisedCSV.to_csv(self.folder + "/supervised.csv", index=False)
        elif self.problem == "unsupervised":
            data = {'images': imagesRoutesArr}
            unsupervisedCSV = pd.DataFrame(data=data)
            unsupervisedCSV.to_csv(self.folder + "/unsupervised.csv", index=False)
        elif self.problem == "regression":
            data = {'images': imagesRoutesArr, 'values': labels}
            regressionCSV = pd.DataFrame(data=data)
            regressionCSV.to_csv(self.folder + "/regression.csv", index=False)

        return image_data, samples

    def _fitAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        # Check if the dimensions are correct ( Attributes => Scale[n,m].size )
        numPixels = self.scale[0] * self.scale[1]
        numAttributes = x.shape[1]
        if numAttributes > numPixels:
            Exception(f"Error: Attributes can't be wrapped in {self.scale}, scale. Please user higher scale.")

        X = x.values
        Y = y.values if y is not None else None

        ranking_feature, corr = self.__generate_feature_distance_ranking(data=X)
        self.coordinate, ranking_image = self.__generate_matrix_distance_ranking(num_r=self.scale[0], num_c=self.scale[1], num=X.shape[1])
        self.index, err, time = self.__training(source=ranking_feature, target=ranking_image)

        self.min_id = np.argmin(err)

    def _transformAlg(self, x: pd.DataFrame, y: Union[pd.DataFrame, None]):
        # Check if the dimensions are correct ( Attributes => Scale[n,m].size )
        numPixels = self.scale[0] * self.scale[1]
        numAttributes = x.shape[1]
        if numAttributes > numPixels:
            Exception(f"Error: Attributes can't be wrapped in {self.scale}, scale. Please user higher scale.")

        X = x.values 
        Y = y.values if y is not None else None

        self.__generate_image_data(
            data=X,
            index=self.index[self.min_id, :],
            num_row=self.scale[0],
            num_column=self.scale[1],
            coord=self.coordinate,
            labels=Y
        )


import pickle
import os
import matplotlib
import pandas as pd
import numpy as np
class BaseModel:
    default_verbose = False  # Verbose: if it's true, show the compilation text
    def __init__(self, verbose=default_verbose):
        self.verbose = verbose

    def saveHyperparameters(self, filename='objs'):
        """
        This function allows SAVING the transformation options to images in a Pickle object.
        This point is basically to be able to reproduce the experiments or reuse the transformation
        on unlabelled data.
        """
        with open(filename+".pkl", 'wb') as f:
            pickle.dump(self.__dict__, f)
        if self.verbose:
            print("It has been successfully saved in " + filename)

    def loadHyperparameters(self, filename='objs.pkl'):
        """
        This function allows LOADING the transformation options to images from a Pickle object.
        This point is basically to be able to reproduce the experiments or reuse the transformation
        on unlabelled data.
        """
        with open(filename, 'rb') as f:
            variables = pickle.load(f)
        
        for key, val in variables.items():
            setattr(self, key, val)

        if self.verbose:
            print("It has been successfully loaded from " + filename)


    def __imageSampleFilter(self, X, Y, coord, matrix, folder):
        """
        This function creates the samples, i.e., the images. This function has the following specifications:
        - The first conditional performs the pre-processing of the images by creating the matrices.
        - Then the for loop generates the images for each sample. Some assumptions have to be taken into
          account in this step:
            - The samples will be created according to the number of targets. Therefore, each folder that is
              created will contain the images created for each target.
            - In the code, the images are exported in PNG format; this can be changed to any other format.
        """
        # Hyperparams
        amplification = self.amplification
        distance = self.distance
        steps = self.steps
        option = self.option

        # Generate the filter
        if distance * steps * amplification != 0:  # The function is only called if there are no zeros (blurring).
            filter = self.__createFilter()

        # In this part, images are generated for each sample.
        for i in range(X.shape[0]):
            matrix_a = np.zeros(matrix.shape)
            if distance * steps * amplification != 0:  # The function is only called if there are no zeros (blurring).
                matrix_a = self.__blurringFilter(matrix_a, filter, X[i], coord)
            else:  # (no blurring)
                iter_values_X = iter(X[i])
                for eje_x, eje_y in coord:
                    matrix_a[int(eje_x), int(eje_y)] = next(iter_values_X)

            extension = 'png'  # eps o pdf
            subfolder = str(int(Y[i])).zfill(2)  # subfolder for grouping the results of each class
            name_image = str(i).zfill(6)
            route = os.path.join(folder, subfolder)
            route_complete = os.path.join(route, name_image + '.' + extension)
            if not os.path.isdir(route):
                try:
                    os.makedirs(route)
                except:
                    print("Error: Could not create subfolder")
            matplotlib.image.imsave(route_complete, matrix_a, cmap='binary', format=extension)

        return matrix


    def __createImage(self, X, Y, folder='prueba/', train_m=False):
        """
        This function creates the images that will be processed by CNN.
        """

        X_scaled = self.min_max_scaler.transform(X)
        Y = np.array(Y)
        try:
            os.mkdir(folder)
            if self.verbose:
                print("The folder was created " + folder + "...")
        except:
            if self.verbose:
                print("The folder " + folder + " is already created...")

        self.m = self.__imageSampleFilter(X_scaled, Y, self.pos_pixel_caract, self.m, folder)

    def __trainingAlg(self, X, Y, folder='img_train/'):
        """
        This function uses the above functions for the training.
        """
        self.__obtainCoord(X)
        self.__areaDelimitation()
        self.__matrixPositions()
        self.__createImage(X, Y, folder, train_m=True)

    def generateImages(self,data, folder):
        """
            This function generate and save the synthetic images in folders.
                - data : data CSV or pandas Dataframe
                - folder : the folder where the images are created
        """
        # Blurring verification

        # Read the CSV
        if type(data) == str:
            dataset = pd.read_csv(data)
            array = dataset.values
        elif isinstance(data,pd.DataFrame) :
            array = data.values
        X = array[:, :-1]
        Y = array[:, -1]

        # Training
        self.__trainingAlg(X, Y, folder=folder)



# -*- coding: utf-8 -*-
"""
Created on Sat Aug 10 18:04:36 2019

@author: Ruibzhan
"""


from mpi4py import MPI 
import paraHill
import pickle
import numpy as np
from itertools import product
import time
import datetime
import config
from config import *
#%% Comm set
comm = MPI.COMM_WORLD
my_rank = comm.Get_rank()
n_processors = comm.Get_size()
print("Processors found: ",n_processors)
current_time = datetime.datetime.now()
print("Time now at the beginning is: ", current_time)

def scatter_list_to_processors(comm, data_list, n_processors):
    import math
    data_amount = len(data_list)
    heap_size = math.ceil(data_amount/(n_processors-1))

    for pidx in range(1,n_processors):
        try:
            heap = data_list[heap_size*(pidx-1):heap_size*pidx]
        except:
            heap = data_list[heap_size*(pidx-1):]
        comm.send(heap,dest = pidx)

    return True

def receive_from_processors_to_dict(comm, n_processors):
    # receives dicts, combine them and return
    feedback = dict()
    for pidx in range(1,n_processors):
        receved = comm.recv(source=pidx)
        feedback.update(receved)
    return feedback

#start = time.time()
#%% load data
with open(args.init,'rb') as file:
    gene_names,dist_matr,init_map = pickle.load(file)
# with open('./Synthetic/Init_Synth8000.pickle','rb') as file:
   # gene_names,dist_matr,init_map = pickle.load(file)
Nn = len(gene_names)

kk = args.num

if init_map.shape[0] != init_map.shape[1]:
    raise ValueError("For now only square images are considered.")
nn = init_map.shape[0]
# Convert from 'F34' to int 34
init_map = np.char.strip(init_map.astype(str),'F').astype(int)
map_in_int = init_map   
#%%
corr_evol = []
if my_rank == 0:
    print("Initial corr: >>>",paraHill.universial_corr(dist_matr,map_in_int))
    for n_iter in range(kk):
        # 9 initial coordinates. 
        init_coords = [x for x in product([0,1,2],repeat = 2)]
        for init_coord in init_coords:
            # Update the mapping. 
            broadcast_msg = map_in_int  
            comm.bcast(broadcast_msg,root = 0)
            # generate the centroids
            xxx = [init_coord[0]+i*3 for i in range(int(nn/3)+1) if (init_coord[0]+i*3)<nn]
            yyy = [init_coord[1]+i*3 for i in range(int(nn/3)+1) if (init_coord[1]+i*3)<nn]
            centr_list = [x for x in product(xxx,yyy)]
            # Master send and recv
            scatter_list_to_processors(comm,centr_list,n_processors)
            swap_dict = receive_from_processors_to_dict(comm,n_processors)
            print(swap_dict)
            map_in_int = paraHill.execute_dict_swap(swap_dict, map_in_int)
            
            print(">",init_coord,"Corr:",paraHill.universial_corr(dist_matr,map_in_int))

        print(">>>",n_iter,"Corr:",paraHill.universial_corr(dist_matr,map_in_int))
        corr_evol.append(paraHill.universial_corr(dist_matr,map_in_int))
        
    coords = np.array([[item[0] for item in np.where(map_in_int == ii)] for ii in range(Nn)])
    with open(args.mapping,'wb') as file:
        pickle.dump([gene_names,coords,map_in_int],file)
    #print("Consumed time:",start - time.time())
    Endtime = datetime.datetime.now()
    print("Time now at the end: ", Endtime)
    import pandas as pd
    pd.Series(corr_evol).to_csv(args.evolution)    
else:
    # other processors
    for n_iter in range(kk):
        broadcast_msg = init_map    # just for a size

        # 9 initial Centroids
        for ii in range(9):
            #Update the mapping
            map_in_int = comm.bcast(broadcast_msg,root = 0)
            
            centr_list = comm.recv(source = 0)
            each_swap_dict = paraHill.evaluate_centroids_in_list(centr_list,dist_matr,map_in_int)
            comm.send(each_swap_dict,dest = 0)
    #result = dict()
    #for each in data:
    #    result.update({each: -each})
    #comm.send(result,dest = 0)

MPI.Finalize



# -*- coding: utf-8 -*-
"""
Created on Thu Oct  8 17:04:34 2020

@author: obazgir
"""

from argparse import ArgumentParser

parser=ArgumentParser()

parser.add_argument('--init', type=str, default='Init.pickle')
parser.add_argument('--mapping', type=str, default='theMapping.pickle')
parser.add_argument('--evolution', type=str, default='Evolv.csv')
parser.add_argument('--num', type=int, default=5)

args= parser.parse_args()


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 12 14:30:06 2019

@author: Ruibzhan & Omid Bazgir
"""

from scipy.stats import pearsonr
import numpy as np
import random
from scipy.spatial import distance
import pickle
import pandas as pd
import time
from itertools import product

#%%
def universial_corr(dist_matr, mapping_in_int):
    # dist_matr is a sqr matr
    Nn = dist_matr.shape[0]
    # find what is the int coordinates for each feature, get an array
    # Because np.where returns a tuple (x_position array, y_position array), a generation is used
    coord = np.array([[item[0] for item in np.where(mapping_in_int == ii)] for ii in range(Nn)])
    # get a 1-d form distance the euclidean dist between pixles positions
    pixel_dist = distance.pdist(coord)
    pixel_dist = pixel_dist.reshape(len(pixel_dist),1)
    # convert the 2-d distance to 1d distance
    feature_dist = distance.squareform(dist_matr)
    feature_dist = feature_dist.reshape(len(feature_dist),1)
    ## pearsonr returns a tuple
    #corrs = pearsonr(feature_dist,pixel_dist)[0]
    L2_Norm = np.sqrt(sum((pixel_dist - feature_dist)**2)/sum(feature_dist**2))
    return L2_Norm
#%%
def evaluate_swap(coord1,coord2,dist_matr,mapping_in_int,original_corr = -2):
    # Coord are in list[]
    # Avoid messing up with the origianl map
    # The correlation before swap can be passed to save some calculation
    the_map = mapping_in_int.copy()
    # If out of bound, return NaN. 
    if coord1[0]<0 or coord1[1]<0 or coord2[0]<0 or coord2[1]<0:
        return np.nan
    if coord1[0]>=the_map.shape[0] or coord1[1]>=the_map.shape[0] or coord2[0]>=the_map.shape[0] or coord2[1]>=the_map.shape[0]:
        return np.nan
    # If not given, recompute.
    if original_corr<-1 or original_corr>1:
        original_corr = universial_corr(dist_matr,the_map)
    # Swap
    try:
        temp = the_map[coord1[0],coord1[1]] 
        the_map[coord1[0],coord1[1]] = the_map[coord2[0],coord2[1]]
        the_map[coord2[0],coord2[1]] = temp
        changed_corr = universial_corr(dist_matr,the_map)
        return float(changed_corr - original_corr)
    except IndexError:
        raise Warning ("Swap index:", coord1,coord2,"Index error. Check the coordnation.")
        return np.nan
    
def evaluate_centroid(centroid,dist_matr,mapping_in_int):
    original_corr = universial_corr(dist_matr,mapping_in_int)
    results = [100000] # just to skip the 0 position
    for each_direc in product([-1,0,1],repeat = 2):
        #print(each_direc)
        # directions are returned as tuple (-1,1), (-1,0), (-1,1), (0,0), ....
        swap_coord = [centroid[0]+each_direc[0],centroid[1]+each_direc[1]]
        evaluation = evaluate_swap(centroid,swap_coord,dist_matr,mapping_in_int,original_corr)
        results.append(evaluation)
    results_array = np.array(results)
    #best_swap_direc = np.where(results_array == np.nanmax(results_array))[0][0]
    best_swap_direc = np.where(results_array == np.nanmin(results_array))[0][0]
    # Give the best direction as a int
    return best_swap_direc

def evaluate_centroids_in_list(centroids_list,dist_matr,mapping_in_int):
    # and returns a dict
    results = dict()
    for each_centr in centroids_list:
        each_centr = tuple(each_centr)
        evaluation = evaluate_centroid(each_centr,dist_matr,mapping_in_int)
        results.update({each_centr:evaluation})
    return results

#%%
def execute_coordination_swap(coord1,coord2,mapping_in_int):
    # try passing the ref. directly 
    the_map = mapping_in_int#.copy()
    # If out of bound, return NaN. 
    if coord1[0]<0 or coord1[1]<0 or coord2[0]<0 or coord2[1]<0:
        raise Warning("Swapping failed:",coord1,coord2,"-- Negative coordnation.")
        return the_map
    if coord1[0]>the_map.shape[0] or coord1[1]>the_map.shape[0] or coord2[0]>the_map.shape[0] or coord2[1]>the_map.shape[0]:
        raise Warning("Swapping failed:",coord1,coord2,"-- Coordnation out of bound.")
        return the_map

    temp = the_map[coord1[0],coord1[1]] 
    the_map[coord1[0],coord1[1]] = the_map[coord2[0],coord2[1]]
    the_map[coord2[0],coord2[1]] = temp

    return(the_map)

# Initial centriod id & Swapping direction: 
# 1 2 3
# 4 5 6
# 7 8 9
# 0 in swapping is preserved for the header.

def execute_direction_swap(centroid,mapping_in_int,direction = 5):
    # Need to notice that [0] is the vertival coord, [1] is the horiz coord. similar to the matlab images.
    coord1 = list(centroid)
    coord2 = list(centroid)
    if direction not in range(1,10):
        raise ValueError("Invalid swapping direction.")
    if direction == 5:
        return mapping_in_int

    if direction in [1,4,7]:
        coord2[1] -=1
    elif direction in [3,6,9]:
        coord2[1] +=1

    if direction in [1,2,3]:
        coord2[0] -=1
    elif direction in [7,8,9]:
        coord2[0] +=1

    the_map = execute_coordination_swap(coord1,coord2,mapping_in_int)
    return the_map

def execute_dict_swap(swapping_dict, mapping_in_int):
    for each_key in swapping_dict:
        execute_direction_swap(each_key,mapping_in_int,direction = swapping_dict[each_key])
    return mapping_in_int   
